###############################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: loader.py,v 1.2 2008-11-26 08:31:33 moscicki Exp $
#
# Copyright (C) 2003-2007 The Ganga Project
#
# This file is part of Ganga. 
#
# Ganga is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Ganga is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
################################################################################

from __future__ import generators
import sys
import os
import os.path
import unittest
from pytf.lib import *
from GangaTest.Framework.tests import GangaGPIPTestCase

__version__ = "1.2"
__author__="Adrian.Muraru[at]cern[dot]ch"


#default timeout in seconds for a test
DEFAULT_TEST_TIMEOUT = 120

class GangaTestLoader:
    """
    Ganga specific TestLoader:
    Loads all the tests from project, filtering based on the given pattern (e.g. Ganga.test.Bugs.* .
    There are two types of tests that may be loaded :
        - *Release Tests* : tests distributed within Ganga distribution and located in a separate branch for each top-level package ( e.g. Ganga/tests/)
        - *Local Tests*   : local test-cases loaded from arbitrary locations 
    """    
    def __init__(self,args,logger=None):        
        
        #optional logger
        self.logger=logger
        
        #test suite files
        self.config = loadArg(args,0, default="default.ini").split(":")
        self.configAlias=''
        
        # automatically append user and site specific configuration when running test-cases        
        from Ganga.Utility.Config import getConfig
        syscfg = getConfig('System')
        self.userConfig = ":".join([os.path.expanduser(syscfg['GANGA_CONFIG_FILE']),
                                    syscfg['GANGA_CONFIG_PATH']])
        
        self.testID = getConfig('Configuration')['user']
        #default path to search tests in
        self.testsTopDir = os.getcwd()
        
        #Ganga top dir
        self.releaseTopDir = os.path.abspath(loadArg(args,1))    
        
        #release testing mode
        self.releaseTesting = int(loadArg(args,2,default=False))
        
        #the path to save the tests' stdout        
        self.outputPrefix = os.path.abspath(loadArg(args,3,default='./'))
        
        self.enableLocalTests = int(loadArg(args,4,default=True))
        self.enableReleaseTests = int(loadArg(args,5,default=True))
        self.report_outputpath = loadArg(args,6,default='./')
        self.schema_test = loadArg(args,7,default='')
        self.current_report_name = None
        
        #set the verbosity based on GangaTest.Framework logging config           
        from Ganga.Utility.logging import getLogger
        tfLogger = getLogger(name='GangaTest.Framework')
        if tfLogger.level < 20 :#DEBUG
            self.verbose = True
        else:
            self.verbose = False
        print self.releaseTesting

    def loadTests(self,patterns):
        """
        pattern format should be: TopLevelPackage.test.A.B.C ... 
        Update 27.02.07:
        -pattern can also be a file/dir containing tests (load tests from arbitrary locations)
        """
        #TODO: support also multiple patterns to be specified ?
        patterns=patterns[:1]
        _log(self.logger,'info',"Test search pattern: %s" % patterns[0])
        
        # release tests take precedence
        if self.enableReleaseTests:
            #work on copy
            _patterns = patterns[:]            
            self.testsTopDir = "%s/python" % self.releaseTopDir
            #_log(self.logger,'info',"Searching release tests in %s " % self.testsTopDir)
            
            new_pattern = self.__convertSearchPattern(_patterns)
            if new_pattern:
                #change the pattern used to select release tests 
                _patterns[0]=new_pattern
            
            pattern = _patterns[0]
            _log(self.logger,'info',"Searching release tests matching [%s] in %s " % (pattern,self.testsTopDir))
            atoms = pattern.split(".")
            print str( patterns ) + " : " + str(atoms)
            if len(atoms) > 2:
                top_level_package = atoms[0]
                tests_type = atoms[1]
                if tests_type == "RT" or tests_type == "test" :
                    release_tests = self._loadReleaseTests(pattern)
                    if release_tests:
                        return release_tests
            #no release test found
            _log(self.logger,'info',"No release tests was found matching %s." % pattern)

        #search for local tests    
        if self.enableLocalTests:
            local_tests = None
            _log(self.logger,'info',"Searching local tests in: %s" % ' '.join(patterns))
            #change the default (release tests) top dir
            self.testsTopDir = os.getcwd()

            local_tests = self._loadLocalTests(patterns)
            if local_tests:
                _msg = "Running local tests from:"
                for p in patterns:
                    _msg = "%s\t %s" % (_msg,p)
                _log(self.logger,'info', _msg)
                return {'unittest':unittest.TestSuite(local_tests)}
            #else
            _log(self.logger,'info',"No local test found...")

        return {}        
    
    def filterTests(self,testcases,pattern):
        import fnmatch
        pattern = pattern.replace('.','/')
        filtered_tests = []        
        for key in testcases.keys():
            tests=testcases[key]            
            for t in tests:
                if key.startswith(pattern) or fnmatch.fnmatch(t.getDescription().split()[0],pattern):
                    filtered_tests.append(t)
        return filtered_tests
                
    def _loadReleaseTests(self,pattern):
        import os,re
        from os.path import join, getsize
        
        top_level_package = pattern.split(".")[0]
        dirname = os.path.abspath("%s/%s/%s"%(self.testsTopDir,top_level_package,"test"))          
        #print "Project RELEASE TOP_DIR: %s"%self.releaseTopDir
        #print "Project TESTS TOP_DIR: %s"%self.testsTopDir
        #print "Package: %s\nPattern: %s "%(top_level_package, pattern)
        print "Test configuration(s):",self.config 

        suites = {}
        tests_no=0
        for config in self.config:
            config_path = "%s/%s/test/config/%s"%(self.testsTopDir,top_level_package,config)
            config_path = "%s:%s" % (config_path,self.userConfig)            
            
            self.configAlias = os.path.splitext(config)[0]
            report_name = "%s__%s"%(pattern.replace("*","ALL"),self.configAlias)
            if self.schema_test is not '':
                report_name = "%s_%s_%s__%s"%(pattern.replace("*","ALL"),self.schema_test,self.testID,self.configAlias)
            print "Using configuration: %s" % config_path 
            print "Report ID: %s" % report_name

            self.current_report_name = report_name

            #the tests container
            tests = self.__walkTests(dirname,pattern,config_path)
            tests_no+=len(tests)
            if tests:
                suites[report_name]=unittest.TestSuite(tests)
        print "%s tests found" % tests_no
        return suites
                
    def __convertSearchPattern(self,patterns):
        """
         translate the input pattern into the internal, dot separated,representation 
        """
        sys_pattern = patterns[0]
        sys_tests_path= os.path.join(self.testsTopDir,sys_pattern)  
        if os.path.exists(sys_tests_path):
            if os.path.isdir(sys_tests_path):
                return os.path.normpath(sys_pattern).replace('/','.')+".*"
            elif os.path.isfile(sys_tests_path):
                if sys_pattern.endswith('.gpi'):
                    return sys_pattern[:-4].replace('/','.')
                elif sys_pattern.endswith('.gpim'):
                    return sys_pattern[:-5].replace('/','.')
                elif sys_pattern.endswith('.py'):
                    return sys_pattern[:-3].replace('/','.')+".*"
                elif sys_pattern.endswith('.gpip'):
                    return sys_pattern[:-5].replace('/','.')+".*"
        return sys_pattern.replace('/','.')
        
        
    def __walkTests(self,dirname,pattern,config):
        """
        Collects all tests from path specified by pattern 
        Parameters:
        pattern : specifies a pattern to select the set of tests to be loaded
        Individual tests can be defined either as separate files (.gpim,.gpi) grouped in directories
        or can be defined in "classic" pyunit test cases (.py)
        The pattern is interpreted as a path to a list of individual tests:
        Example:
        The following physical structure for tests:
            Ganga/test/
                  /Folders/
                      testA.gpi
                      testB.gpi
                      TestGroup.py {multiple tests embedded: testC, testD...}
                  /Folders.py {multiple tests embedded: testE, testF...}
        
        is transformed in the logical structure:
            Ganga/tests/Folders
                           /{testA,testB,testE,testF}
                           /TestGroup{testC,testD}                     
        So, the pattern Ganga.test.Folders.* will match all the tests 
        while the pattern Ganga.test.Folders.TestGroup.* will match just the test belonging to TestGroup
        Invidual tests can be selected by specifying the complete path: 
                  Ganga.test.Folders.TestGroup.testE or Ganga.test.Folders.testA
        
        N.B.:
            It may be possible that a complete path to match multiple tests if a test with the same name is defined as a single test(.gpi,.gpim) in a directory
            and also in a PyUnit testcase class
            
        """

        tests={}
        #the tests container
        topdir = self.testsTopDir        
        for root,dirs,files in walk(os.path.join(dirname)):
            if len(files)>0:
                for file in files:
                    #if file.startswith('.'): print "Skipped", file; continue;
                    if file.endswith(".gpi") or file.endswith(".gpim"):
                        key=root[root.index(topdir)+len(topdir)+1:]
                        list = tests.get(key,[])
                        if file.endswith(".gpi"):
                            test = self.__generateTestCaseFromGPITest(os.path.join(root,file),config)
                            if test is not None: list.append(test)
                        elif file.endswith(".gpim"):
                            test = self.__generateTestCaseFromGMPTest(os.path.join(root,file),config)
                            if test is not None: list.append(test)
                        tests[key]=list
                    elif file.endswith(".py") and file.startswith('Test'):
                        newroot= os.path.join(root,get_name(file))
                        key=newroot[newroot.index(topdir)+len(topdir)+1:]
                        list = tests.get(newroot,[])
                        #list.append("%s tests"%file)
                        pys = self.__generateTestCasesFromPyUnitTest(os.path.join(root,file),config)
                        if type(pys)==type([]) and len(pys)>0:
                            list+=pys
                        tests[key]=list       
                    elif file.endswith(".gpip") and file.startswith('Test'):
                        newroot= os.path.join(root,get_name(file))
                        key=newroot[newroot.index(topdir)+len(topdir)+1:]
                        list = tests.get(newroot,[])
                        #list.append("%s tests"%file)
                        gpips = self.__generateTestCasesFromGPIPTest(os.path.join(root,file),config)
                        if type(gpips)==type([]) and len(gpips)>0:
                            list+=gpips
                        tests[key]=list       
        filtered = self.filterTests(tests,pattern)
        filtered.sort(lambda t1,t2: cmp(t1.getDescription(),t2.getDescription()))
        return filtered

    def _loadLocalTests(self,locations):
        """
        Load tests from *location* : can be a single file or a directory
        """
        import os,glob
        tests = []
        for location in locations:            
            if os.path.exists(location):
                paths = [location]
            paths = glob.glob(location)            
            if len(paths)==0:
                guess = [location+".gpi",location+".gpim",location+".py",location+".gpip"]                
                paths = [l for l in guess if os.path.exists(l) and os.path.isfile(l)]
                
            if len(paths)==0:
                continue
            
            #else, we found smth
            for l in paths:                
                #if the location is inside GANGA_PACKAGE/test/ dir then
                #use (as in release tests) the GANGA_PACKAGE/test/config dir to load
                #the configuration from
                tpath = fullpath(l)
                rpath = os.path.join(self.releaseTopDir,'python')
                idx = tpath.rfind(rpath)                
                if idx >= 0:
                    package = tpath[idx+len(rpath)+1:].split(os.sep)[0]
                else:
                    package = 'Ganga'                

                if os.path.isfile(l):
                    tests+=self.__loadLocalTest(l,gangaPackage=package)
                elif os.path.isdir(l):
                    for file in dirwalk(l):
                        tests+=self.__loadLocalTest(file,gangaPackage=package)
        if len(tests)==0:
            return None        
        tests.sort(lambda t1,t2: cmp(t1.getDescription(),t2.getDescription()))
        return tests
    
    def __loadLocalTest(self,file,gangaPackage='Ganga'):
        """
        Load test(s) from file
        Return:
         a list containing PYTF testcases
        """
        list = []
        
        # for each test configuration specified 
        for config in self.config:              
            self.configAlias = os.path.splitext(config)[0]            
            # construct the test configuration path
            # 1. User config : GANGA_CONFIG_FILE:GANGA_CONFIG_PATH
            config_path = "%s/python/%s/test/config/%s"%(self.releaseTopDir,gangaPackage,config)
            ganga_config_path = ":".join([config_path,self.userConfig])

            ext = get_ext(file)
            if ext in (".gpi",".gpim"):
                test = None
                if ext == ".gpi":
                    test = self.__generateTestCaseFromGPITest(file,ganga_config_path,test_relpath=os.path.dirname(file))
                elif ext == ".gpim":
                    test = self.__generateTestCaseFromGMPTest(file,ganga_config_path,test_relpath=os.path.dirname(file))
                if test is not None: list.append(test)
            elif ext.lower() == ".gpip" and get_name(file).startswith('Test'):
                print 'Load local GPIP test'
                gpips = []
                gpips = self.__generateTestCasesFromGPIPTest(file,ganga_config_path,test=None,test_relpath=os.path.dirname(file))
                if type(gpips)==type([]) and len(gpips)>0:
                    list+=gpips
            elif ext.lower() == ".py" and get_name(file).startswith('Test'):            
                pys=[]
                pys = self.__generateTestCasesFromPyUnitTest(file,ganga_config_path,test=None,test_relpath=os.path.dirname(file))
                if type(pys)==type([]) and len(pys)>0:
                    list+=pys
                
        return list


    def __generateTestCaseFromCmd(self, testCmd, testName, testAlias=None, 
                                setUpCmd=None, tearDownCmd=None, description=None, 
                                output_path=None, timeout=sys.maxsize):
        """
        Add a single ganga test-case from a gpi test (similar to run_external_test)
        Internally this method decorates the setUp->[testCmd]->tearDown in a TestCase
        Arguments:
         output_path: when specified, then the output is captured in the given file
         timeout: the testcase is reported as failed if it passes the specified time in seconds
        """
        if testAlias is None:
            testAlias = testName
        method_name = testName
        d = {}
        #todo: use a VARS dictionary for variable used to create the wrapper script
        VARS={}
        VARS['method_name']=method_name
        VARS['verbose'] = self.verbose
        VARS['timeout'] =timeout
        VARS['output_path']=output_path
        VARS['description']=description
        VARS['setUpCmd']=setUpCmd
        VARS['tearDownCmd']=tearDownCmd    
        
        code="""

def %(method_name)s(self):
    import unittest
    import time
    
    def read_file(filename):
        f = open(filename)
        try: return f.read()
        finally: f.close()

    def write_file(filename,content):
        f = open(filename,'w')
        try: return f.write(content+"\\n")
        finally: f.close()
   
    process = None # used as handler to the test process    
    pytf_paths = str.split(os.getenv('PYTF_TOP_DIR'),':')
    pytf_runner = str()
    if type(pytf_paths) is str:
        pytf_runner = os.path.join( pytf_paths, 'cmd_run.sh' )
    else:
        for i in pytf_paths:
            if os.path.isfile( os.path.join( i, 'cmd_run.sh') ):
                pytf_runner = os.path.join( i, 'cmd_run.sh')
                break;
    #pytf_runner = os.path.join(os.getenv('PYTF_TOP_DIR',''),'cmd_run.sh')    
    pytf_runner = os.path.join(os.getenv('PYTF_TOP_DIR','').split(':')[0],'cmd_run.sh')

    def getstatusouterr(cmd,output_path="%(output_path)s",out_mode='a',setTimeout=False):
        import os
        import os.path
        import time
        from subprocess import PIPE,STDOUT,Popen        
        info = ''
        if not os.path.exists(pytf_runner):
            print '\\n ERROR: Cannot find %%s\\n' %% pytf_runner
        script_runner = [pytf_runner]
        if setTimeout:
            script_runner.extend(['-t','%(timeout)s'])
        else:
           #hard value for timeout (1h)
           script_runner.extend(['-t','3600'])

        try:
            import Ganga.Core.InternalServices.Coordinator
            if Ganga.Core.InternalServices.Coordinator.servicesEnabled:
                from Ganga.GPI import disableServices
                disableServices()
        
            from Ganga.GPI import  reactivate
            reactivate()
        
            from Ganga.GPI import jobs, templates
            for j in jobs: j.remove()
            for t in templates: t.remove()
            if hasattr(jobs,'clean'):
                jobs.clean(confirm=True, force=True)
            if hasattr(templates,'clean'):
                templates.clean(confirm=True, force=True)
        
            from Ganga.GPI import disableServices
            disableServices()
        
        #    # DANGEROUS only use as a last resort to files being kept open
        #    #from Ganga.Runtime import bootstrap
        #    #bootstrap.safeCloseOpenFiles()
        #
        except:
            pass

        #try:
        #    from Ganga.GPI import disableServices
        #    disableServices()
        #except:
        #    pass

        script_runner.append('{ ' + cmd + '; }')
        output = open(output_path,out_mode)
        output.write('########## Test started: ' + time.ctime() + ' ##########')
        output.flush()
        print "Running: \\n" + str(script_runner) + "\\n"
        process = Popen(script_runner, shell=False, bufsize=0, 
                        stdin=PIPE, stdout=output, stderr=STDOUT, close_fds=True)        
        process.wait()
        output.write('########## Test finished: ' + time.ctime() + ' ##########')
        output.write("\\n")
        output.close()
        sts = process.returncode
        if sts is None:
           sts = 0
        #if timeout
        if sts == 7:
           raise unittest.TestCase.failureException("Test Timeout(%(timeout)s sec). " 
                      "Consider increasing the value of '[TestingFramework]timeout' parameter in your test configuration file")

        text = read_file(output_path)        
        tb_index = text.rfind('Traceback')
        out=text[:tb_index]
        err=text[tb_index:]

        #try:
        #    import Ganga.Core.InternalServices.Coordinator
        #    if Ganga.Core.InternalServices.Coordinator.servicesEnabled:
        #        from Ganga.GPI import disableServices
        #        disableServices()
        #
        #    from Ganga.GPI import reactivate
        #    reactivate()
        #
        #
        #    from Ganga.GPI import jobs, templates
        #    for j in jobs: j.remove()
        #    for t in templates: t.remove()
        #    if hasattr(jobs,'clean'):
        #        jobs.clean(confirm=True, force=True)
        #    if hasattr(templates,'clean'):
        #        templates.clean(confirm=True, force=True)
        #
        #    from Ganga.GPI import disableServices
        #    disableServices()
        
        #except:
        #    pass

        #try:
        #    from Ganga.GPI import reactivate
        #    reactivate()
        #except:
        #    pass

        return sts,out,err
        
    #delete the output file if it already exists
    try:
        os.unlink("%(output_path)s")
    except OSError as e:
        pass
    
""" % VARS
        
        code=code+"""
    try:
        print "\\n****** %(description)s"
        print "\\tLog file: %(output_path)s"
        import signal
"""%VARS
        
        if setUpCmd is not None:
            code=code+"""
        import sys
        print "\\tSetup test..."
        (status,out,err)=getstatusouterr('%(setUpCmd)s')
        if status: 
            print "\\tSetup failed: ",out,' ',err
        else: 
            print "\\tSetup done."
"""%VARS
        
        for cmd in testCmd:
            code=code+"""
        (status,out,err)=getstatusouterr('%s',setTimeout=True)
        if status: raise unittest.TestCase.failureException, err;
"""%cmd
        code=code+"""                
    finally:
        #cleanup:
        pass
          
"""
        if tearDownCmd is not None:
            code=code+"""
        import sys
        (status,out,err)=getstatusouterr('%(tearDownCmd)s')
        if status: 
            print "Cleanup failed: ",out,' ',err        
"""%VARS
        else:
            code=code+"""
        pass
"""
        
        
        #print code
        exec(code, globals(), d)
        __method = d[method_name].__get__(self, self.__class__)
        setattr(self, method_name, __method)
        return NamedFunctionTestCase(__method,description=description)

    def __generateTestCaseFromGPITest(self,test_path,ganga_config_path,test_relpath=None):
        """
        test_path: relative test path to self.testsTopDir
        """
        test_filepart = os.path.basename(test_path)
        test_dir = os.path.dirname(test_path)   
        test_filename = os.path.splitext(test_filepart)[0]
        
        # test specific configuration file
        test_ini = os.path.join(self.testsTopDir,test_dir,test_filename+".ini")
        # test setup script 
        test_setup = os.path.join(self.testsTopDir,test_dir,test_filename+".setup")
        # test tear down command
        test_cleanup = os.path.join(self.testsTopDir,test_dir,test_filename+".cleanup")

        if test_relpath is None: #relative to release
            test_relpath = test_path[len(self.testsTopDir)+1:-len(os.path.splitext(test_filepart)[1])]
            
        test_name = "%s (GPI)"%test_relpath
        
        if self.releaseTesting:
            #when doing release testing we generate the coverage analysis reports
            #and generatate a well defined output dir structure
            if self.schema_test is not '':
                output_path=os.path.join(self.outputPrefix,test_relpath.replace("/",".")+"_"+self.schema_test+"_"+self.testID+"_"+ self.configAlias)
            else:
                output_path=os.path.join(self.outputPrefix,test_relpath.replace("/",".")+"__"+ self.configAlias)
            coverage_path = "%s.figleaf" % output_path
        else:
            output_path=os.path.join(self.outputPrefix,test_filename)
            if self.configAlias:
                output_path+="__%s" % self.configAlias
            coverage_path = ""
         
        config = ganga_config_path
        if os.path.isfile(test_ini):
            config = "%s:%s"%(test_ini,config)
    
        #read the timeout for this testcase            
        timeout = readTestTimeout(config)
        
        def opt_path(coverage_path):
            if coverage_path:
                return coverage_path
            return 'None'
        runPyFile = os.path.join(self.releaseTopDir,"python","GangaTest","Framework","driver.py")
        from Ganga.Utility.Config import getConfig
        if self.schema_test is not '':
            newgangadir = os.path.join(getConfig('Configuration')['gangadir'],self.schema_test)
            testCmd ="cd %s ; env --unset=GANGA_INTERNAL_PROCREEXEC OUTPUT_PATH=%s PYTHONUNBUFFERED=1 "\
                    "%s/bin/ganga -o[Configuration]gangadir=%s -o[Configuration]RUNTIME_PATH=GangaTest --config= --config-path=%s "\
                    "%s --test-type=gpi --coverage-report=%s %s"%(
                os.path.join(self.testsTopDir,test_dir),
                fullpath(output_path),
                self.releaseTopDir,            
                newgangadir,
                config, runPyFile,
                opt_path(coverage_path),
                test_path)        
        else:
            testCmd ="cd %s ; env --unset=GANGA_INTERNAL_PROCREEXEC OUTPUT_PATH=%s PYTHONUNBUFFERED=1 "\
                    "%s/bin/ganga -o[Configuration]RUNTIME_PATH=GangaTest --config= --config-path=%s "\
                    "%s --test-type=gpi --coverage-report=%s %s"%(
                os.path.join(self.testsTopDir,test_dir),
                fullpath(output_path),
                self.releaseTopDir,            
                config, runPyFile,
                opt_path(coverage_path),
                test_path)        


        # setup Ganga-Session
        if os.path.isfile(test_setup):
            setUpCmd = "cd %s ; PYTHONUNBUFFERED=1  python %s"%(
                os.path.join(self.testsTopDir,test_dir),                
                test_setup)
        else:
            setUpCmd = None

        # tearDown Ganga Session
        if os.path.isfile(test_cleanup):
            tearDownCmd = "cd %s ; PYTHONUNBUFFERED=1 python %s"%(
                os.path.join(self.testsTopDir,test_dir),                
                test_cleanup)
        else: 
            tearDownCmd = None
            #default cleanup code is located in GangaTest/Framework/cleanup.py and is used by default when running GPI scripts            
            #cleanUpPyFile = os.path.join(self.releaseTopDir,"python","GangaTest","Framework","cleanup.py")
            #tearDownCmd = "cd %s ; env --unset=GANGA_INTERNAL_PROCREEXEC OUTPUT_PATH=%s PYTHONUNBUFFERED=1 "\
            #              "%s/bin/ganga -o[Configuration]RUNTIME_PATH=GangaTest --config= --config-path=%s --quiet %s"%(
            #    os.path.join(self.testsTopDir,test_dir),
            #    fullpath(output_path),                
            #    self.releaseTopDir,
            #    config,
            #    cleanUpPyFile)
        try:
            descr = os.path.join(test_dir[test_dir.index(self.testsTopDir)+len(self.testsTopDir)+1:],test_filename)+" [GPI]"
        except ValueError:
            descr = os.path.join(test_dir,test_filename)+" [GPI]"

        return self.__generateTestCaseFromCmd(
            [testCmd], 
            test_filename,
            testAlias = test_name,
            setUpCmd=setUpCmd,
            tearDownCmd=tearDownCmd,
            description=descr,
            output_path=output_path+'.out',
            timeout=timeout
            ) 
        
    def __generateTestCaseFromGMPTest(self, test_path,ganga_config_path,test_relpath=None):
        
        test_filepart = os.path.basename(test_path)
        test_dir = os.path.dirname(test_path)   
        test_filename = os.path.splitext(test_filepart)[0]
        test_ini_prefix = os.path.join(self.testsTopDir,test_dir,test_filename)
        test_ini = test_ini_prefix + ".ini"

        if test_relpath is None: #relative to release
            test_relpath = test_path[len(self.testsTopDir)+1:-len(os.path.splitext(test_filepart)[1])]
            
        test_name = "%s (GPIM)"%test_relpath
        
        if self.releaseTesting:
            #when doing release testing we generate the coverage analysis reports
            #and generatate a well defined output dir structure
            output_path=os.path.join(self.outputPrefix,test_relpath.replace("/",".")+"__"+ self.configAlias)
            coverage_path = "%s.figleaf" % output_path
        else:
            output_path=os.path.join(self.outputPrefix,test_filename)
            if self.configAlias:
                output_path+="__%s" % self.configAlias
            coverage_path = ""
 
        if os.path.isfile(test_ini):
            test_ini = "%s:%s"%(test_ini,ganga_config_path)
        else:
            test_ini = ganga_config_path

        #read the timeout for this testcase
        timeout = readTestTimeout(test_ini)
        
        #passes may have specific configuration files (e.g. MyTest.pass1.ini)
        test_ini = '%s:'+ test_ini

        #testCmdPrefix = "cd %s ; PYTHONUNBUFFERED=1 PYTHONPATH=%s %s/bin/ganga -o[Configuration]RUNTIME_PATH=GangaTest --config= --config-path=%s"%(
        testCmdPrefix = "cd %s ; env PYTHONUNBUFFERED=1 OUTPUT_PATH=%s "\
                      "%s/bin/ganga -o[Configuration]RUNTIME_PATH=GangaTest --config= --config-path=%s" % \
                        (os.path.join(self.testsTopDir,test_dir), 
                         fullpath(output_path),
                         self.releaseTopDir,
                         test_ini)
        try:
            module = exec_module_code(test_path)
        except Exception as e:
            print "[WARNING] Cannot parse Multipass test (invalid format) %s [skipped]-> %s"%(test_path,e)
            return 
            
        testCmds = []
        runPyFile = os.path.join(self.releaseTopDir,"python","GangaTest","Framework","driver.py")
        for name in dir(module):

            clazz = getattr(module, name)
            clazz_dict = dir(clazz)
            sclazz=str(clazz)
            if isinstance(clazz, (type, types.ClassType)) and issubclass(clazz, MultipassTest) and sclazz.find(test_filename)>=0:
                if "run" not in clazz_dict:
                    print "Cannot find <run> method in the MultipassTest class"
                    break
                #for each multipass tests: build n test-cases
                for attrname in clazz_dict:                   
                    # if is "pass%i" method
                    if attrname.startswith("pass"):
                        pass_ini = '%s.%s.ini' % (test_ini_prefix,attrname)
                        if not os.path.isfile(pass_ini):
                            pass_ini = ''
                        targs = 'run'
                        if attrname=="pass1":
                            targs = 'run --multipass-restart setUp'                     
                        testCmds.append( "%s %s %s --test-type=gpim --coverage-report=%s %s" % \
                                         (testCmdPrefix % pass_ini,
                                          runPyFile,
                                          test_path,
                                          coverage_path,
                                          targs))
                if "tearDown" in clazz_dict and len(testCmds)>0:
                    testCmds[-1] += ' tearDown'
                break
        try:
            descr = os.path.join(test_dir[test_dir.index(self.testsTopDir)+len(self.testsTopDir)+1:],test_filename)+" [GPIM]"
        except ValueError:
            descr = os.path.join(test_dir,test_filename)+" [GPIM]"
        if len(testCmds)>0:
            return self.__generateTestCaseFromCmd(
                testCmds,
                test_filename,
                testAlias = test_name,
                description=descr,
                output_path=output_path+'.out',
                timeout=timeout)
        
    def __generateTestCasesFromPyUnitTest(self,test_path,ganga_config_path,test=None,test_relpath=None):
        
        test_filepart = os.path.basename(test_path)
        test_dir = os.path.dirname(test_path)
        test_filename = os.path.splitext(test_filepart)[0]
        
        if test_relpath is None: #relative to release
            test_relpath = test_path[len(self.testsTopDir)+1:-len(os.path.splitext(test_filepart)[1])]
            
        test_name = "%s"%test_relpath
        
        if self.releaseTesting:
            #when doing release testing we generate the coverage analysis reports
            #and generatate a well defined output dir structure
            output_path=os.path.join(self.outputPrefix,test_relpath.replace("/",".")+"."+"%s"+"__"+self.configAlias+".out")
            coverage_path = "%s.figleaf" % os.path.splitext(output_path)[0]
        else:
            output_path=os.path.join(self.outputPrefix,test_filename+"."+"%s"+".out")            
            if self.configAlias:
                output_path+="__%s" % self.configAlias
            coverage_path = ""

        def getCoveragePath(template,value):
            if template:
                return template%value
            return ''
            
        #print "Checking PyUnit tests from: %s"%test_name
        test_ini = os.path.join(self.testsTopDir,test_dir,test_filename+".ini")

        #set the ganga configuration path
        if os.path.isfile(test_ini):
            test_ini = "%s:%s"%(test_ini,ganga_config_path)
        else:
            test_ini = ganga_config_path

        #read the timeout for this testcase
        timeout = readTestTimeout(test_ini)
        
        try:
            module = exec_module_code(test_path)
            if hasattr(module,'setUp'):
                getattr(module,'setUp')()
        except Exception as e:
            print "[WARNING] Cannot parse PYUnit test  %s [skipped]:"%test_path
            import traceback
            traceback.print_exc()
            return []
            
        tests = []
        runPyFile = os.path.join(self.releaseTopDir,"python","GangaTest","Framework","driver.py")
        for name in dir(module):

            try:
                clazz = getattr(module, name)
                sclazz = str(clazz)
            #print("name: %s" % str(name))
            #print("sclazz: %s" % str(sclazz))
                if isinstance(clazz, (type, types.ClassType)) and  sclazz.split('.')[-1].find(test_filename)>=0:
                    clazz_dict = dir(clazz)
                    for attrname in clazz_dict:
                        try:
                            descr = os.path.join(test_dir[test_dir.index(self.testsTopDir)+len(self.testsTopDir)+1:],test_filename,attrname)+" [PY]"
                        except ValueError:
                            descr = os.path.join(test_dir,test_filename,attrname)+" [PY]"
                        if test is not None and attrname==test:
                            testCmd = "cd %s ; env OUTPUT_PATH=%s "\
                                "%s/bin/ganga -o[Configuration]RUNTIME_PATH=GangaTest --config= --config-path=%s %s %s --test-type=py --coverage-report=%s %s" % (
                            os.path.join(self.testsTopDir,test_dir),
                            fullpath(output_path%attrname),                            
                            self.releaseTopDir,
                            test_ini, runPyFile,
                            test_path, 
                            getCoveragePath(coverage_path,attrname),
                            attrname)

                            testcase = self.__generateTestCaseFromCmd(
                            [testCmd],
                            attrname,
                            testAlias="%s.%s (PY)"%(test_name,attrname),
                            description=descr,
                            output_path=output_path%attrname,
                            timeout=timeout
                            )
                            if testcase is not None: return [testcase]
                        elif attrname.startswith("test") and test is None:
                            #syspath = [p for p in sys.path if p.find("/usr/lib")==-1]                        
                            testCmd = "cd %s ;env OUTPUT_PATH=%s "\
                                "%s/bin/ganga -o[Configuration]RUNTIME_PATH=GangaTest --config= --config-path=%s %s %s --test-type=py --coverage-report=%s %s setUp tearDown"%(
                            os.path.join(self.testsTopDir,test_dir),
                            fullpath(output_path%attrname),
                            self.releaseTopDir,
                            test_ini, 
                            runPyFile,
                            test_path,
                            getCoveragePath(coverage_path,attrname),
                            attrname)                        
                            testcase = self.__generateTestCaseFromCmd(
                            [testCmd],
                            attrname,
                            testAlias="%s.%s (PY)"%(test_name,attrname),
                            description=descr,
                            output_path=output_path%attrname,
                            timeout=timeout
                            )
                            if testcase is not None: tests.append(testcase)
            except:
                pass

                break
        return tests

    def __generateTestCasesFromGPIPTest(self,test_path,ganga_config_path,test=None,test_relpath=None):
        
        test_filepart = os.path.basename(test_path)
        test_dir = os.path.dirname(test_path)
        test_filename = os.path.splitext(test_filepart)[0]
        
        if test_relpath is None: #relative to release
            test_relpath = test_path[len(self.testsTopDir)+1:-len(os.path.splitext(test_filepart)[1])]
            
        test_name = "%s"%test_relpath
        
        if self.releaseTesting:
            #when doing release testing we generate the coverage analysis reports
            #and generatate a well defined output dir structure
            output_path=os.path.join(self.outputPrefix,test_relpath.replace("/",".")+"."+"%s"+"__"+self.configAlias+".out")
            coverage_path = "%s.figleaf" % os.path.splitext(output_path)[0]
        else:
            output_path=os.path.join(self.outputPrefix,test_filename+"."+"%s"+".out")            
            if self.configAlias:
                output_path+="__%s" % self.configAlias
            coverage_path = ""

        def getCoveragePath(template,value):
            if template:
                return template%value
            return ''
            
        #print "Checking PyUnit tests from: %s"%test_name
        test_ini = os.path.join(self.testsTopDir,test_dir,test_filename+".ini")

        #set the ganga configuration path
        if os.path.isfile(test_ini):
            test_ini = "%s:%s"%(test_ini,ganga_config_path)
        else:
            test_ini = ganga_config_path

        #read the timeout for this testcase
        timeout = readTestTimeout(test_ini)
        
        try:
            module = exec_module_code(test_path)
            if hasattr(module,'setUp'):
                getattr(module,'setUp')()
        except Exception as e:
            print "[WARNING] Cannot parse GPIP test  %s [skipped]:"%test_path
            import traceback
            traceback.print_exc()
            return []
            
        tests = []
        runPyFile = os.path.join(self.releaseTopDir,"python","GangaTest","Framework","driver.py")
        for name in dir(module):
            clazz = getattr(module, name)
            sclazz = str(clazz)
            #print sclazz
            #print '-->%s, -->%s' % (sclazz.split('.')[-1].split("'")[0], test_filename)
            if isinstance(clazz, (type, types.ClassType)) and  sclazz.split('.')[-1].split("'")[0] == test_filename:
                #print '[%s]' % test
                if test is not None and attrname==test:
                    print 'This line should be printed out.'
                else:
                    try:
                        descr = os.path.join(test_dir[test_dir.index(self.testsTopDir)+len(self.testsTopDir)+1:],test_filename,'ALL')+" [GPIP]"
                    except ValueError:
                        descr = os.path.join(test_dir,test_filename,'ALL')+" [GPIP]"
                    testCmd = "cd %s ;env OUTPUT_PATH=%s "\
                              "%s/bin/ganga -o[Configuration]RUNTIME_PATH=GangaTest --config= --config-path=%s %s %s --test-type=gpip --output_base=%s --coverage-report=%s --timeout=%s %s setUp tearDown"%(
                              os.path.join(self.testsTopDir,test_dir),
                              fullpath(output_path),
                              self.releaseTopDir,
                              test_ini,
                              runPyFile,
                              test_path,
                              output_path,
                              getCoveragePath(coverage_path,'ALL'),
                              timeout,
                              'run_pyunit_testcases')

                    #print '%s *********' % descr
                    testcase = GangaGPIPTestCase(testCmd, descr,output_path,self.report_outputpath,self.current_report_name,self.releaseTesting)
                    if testcase is not None: tests.append(testcase)
        #print tests
        return tests

## Utility functions ##

def loadArg(args,ind,default=None):
    try:
        if args[ind] == '':
            return default
        else:
            return args[ind]
    except IndexError:
        return default
    
def readTestTimeout(config):
    """get the timeout value from the list of test configuration files (colon separated list)"""
   
    import ConfigParser
    parser = ConfigParser.ConfigParser()
    timeout=DEFAULT_TEST_TIMEOUT
    for file in config.split(':'):
        try:
            parser.read(file)
            timeout = parser.get("TestingFramework","timeout")
            break
        except:
            #not found or error
            continue
    return timeout          

def get_ext(filename):
    return os.path.splitext(os.path.split(filename)[1])[1]

def get_name(filename):
    return os.path.splitext(os.path.split(filename)[1])[0]

def dirwalk(dir):
    "walk a directory tree, using a generator"
    for f in os.listdir(dir):
        fullpath = os.path.join(dir,f)
        if os.path.isdir(fullpath) and not os.path.islink(fullpath):
            for x in dirwalk(fullpath):  # recurse into subdir
                yield x
        else:
            yield fullpath
            
def walk(top, topdown=True, onerror=None):
    """
    os.path.walk function 
    """    
    from os.path import join, isdir, islink
    from os import listdir
    # We may not have read permission for top, in which case we can't
    # get a list of the files the directory contains.  os.path.walk
    # always suppressed the exception then, rather than blow up for a
    # minor reason when (say) a thousand readable directories are still
    # left to visit.  That logic is copied here.
    try:
        names = listdir(top)
    except Exception as err:
        print err
        if onerror is not None:
            onerror(err)
        return

    dirs, nondirs = [], []
    for name in names:
        if isdir(join(top, name)):
            dirs.append(name)
        else:
            nondirs.append(name)

    if topdown:
        yield top, dirs, nondirs
    for name in dirs:
        path = join(top, name)

        if not islink(path):
            for x in walk(path, topdown, onerror):
                yield x
    if not topdown:
        yield top, dirs, nondirs
        
def expandfilename(filename):
    "expand a path or filename in a standard way so that it may contain ~ and ${VAR} strings"
    return os.path.expandvars(os.path.expanduser(filename))

def fullpath(path):
    "expandfilename() and additionally: strip leading and trailing whitespaces and expand symbolic links"
    return os.path.realpath(expandfilename(path.strip()))

def _log(logger,level,msg):
    if logger and hasattr(logger,level):
        getattr(logger,level)(msg)
    else:
        print "[%s] %s" % (level.upper(),msg)

#$Log: not supported by cvs2svn $
#Revision 1.1  2008/07/17 16:41:36  moscicki
#migration of 5.0.2 to HEAD
#
#the doc and release/tools have been taken from HEAD
#
#Revision 1.43.4.6  2008/02/12 14:08:02  amuraru
#*** empty log message ***
#
#Revision 1.43.4.5  2007/12/18 13:12:48  amuraru
#stream-line the test-case execution code (use a single entry-point : Framework/driver.py)
#symplified the interface with Ganga bootstrap
#
#Revision 1.43.4.4  2007/12/17 15:51:29  amuraru
#use subprocess in the Test Runner
#
#Revision 1.43.4.3  2007/12/07 17:30:33  amuraru
#- use the timeout functionality from cmd_run shell wrapper
#
#Revision 1.43.4.2  2007/11/21 12:18:58  amuraru
#fixed the loader imports to work on python 2.5
#
#Revision 1.43.4.1  2007/10/31 13:37:21  amuraru
#updated to the new config subsystem in Ganga 5.0
#
#Revision 1.43  2007/09/13 15:18:50  amuraru
#fixed the local configuration
#
#Revision 1.42  2007/09/11 16:17:06  amuraru
#*** empty log message ***
#
#Revision 1.41  2007/09/11 15:44:59  amuraru
#-delete the test output file if it already exists
#
#Revision 1.40  2007/09/04 09:48:13  amuraru
#restructed code to allow two modes of running: developer mode and release mode
#
#Revision 1.39  2007/08/21 14:38:44  amuraru
#*** empty log message ***
#
#Revision 1.38  2007/08/21 14:30:22  amuraru
#code cleanup
#
#Revision 1.37  2007/08/13 17:20:37  amuraru
#- fix bug #28430
#- do not generate coverage analysis reports for local tests
#- cleanup
#
#Revision 1.36  2007/08/13 12:33:15  amuraru
#added verbosity control, avoid purging the used reporisotry when running local tests
#
#Revision 1.35  2007/06/06 14:27:25  amuraru
#
#minor fixes
#
#Revision 1.34  2007/06/05 23:17:08  amuraru
#*** empty log message ***
#
#Revision 1.33  2007/06/04 13:24:11  amuraru
#*** empty log message ***
#
#Revision 1.32  2007/06/01 09:00:36  amuraru
#small fix in GPIM loading
#
#Revision 1.31  2007/05/31 11:40:03  amuraru
#- updated multi-pass tests loader to read pass specific configuration files (if exist)
# i.e: MyMultipass.pass<no>.ini are searched in the test driectory and added in ganga config-path
#
#Revision 1.30  2007/05/21 16:01:20  amuraru
#use default website css style in test/coverage reports;
#disabled per test-case coverage report generation;
#other fixes
#
#Revision 1.29  2007/05/16 10:15:52  amuraru
#use ganga logger
#
#Revision 1.28  2007/05/15 09:58:36  amuraru
#html reporter updated
#



##############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: AthenaLCGRTHandler.py,v 1.51 2009-07-23 20:12:25 elmsheus Exp $
###############################################################################
# TagPrepare LCG Runtime Handler
#
# ATLAS/ARDA

import os, pwd, commands, re, string, time, sys

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *
from Ganga.GPIDev.Lib.File import *

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from Ganga.Utility.Config import getConfig, makeConfig, ConfigError
from Ganga.Utility.logging import getLogger

from Ganga.Lib.LCG import LCGJobConfig
from GangaAtlas.Lib.AtlasLCGRequirements import AtlasLCGRequirements

from GangaAtlas.Lib.ATLASDataset import ATLASDataset, isDQ2SRMSite, getLocationsCE, getIncompleteLocationsCE, getIncompleteLocations, whichCloud
from GangaAtlas.Lib.ATLASDataset import DQ2Dataset
from GangaAtlas.Lib.ATLASDataset import DQ2OutputDataset
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler

__directory__ = os.path.dirname(__file__)
__athdirectory__ = sys.modules['GangaAtlas.Lib.Athena'].__path__[0]

def _append_file_buffer(inputbox,name,array):

    inputbox.append(FileBuffer(name,'\n'.join(array)+'\n'))

def _append_files(inputbox,*names):

    for name in names:
        inputbox.append(File(os.path.join(__directory__,name)))

class TagPrepareLCGRTHandler(IRuntimeHandler):
    """Athena LCG Runtime Handler"""

    def prepare(self,app,appsubconfig,appmasterconfig,jobmasterconfig):
        """prepare the subjob specific configuration"""

        job = app._getParent() # Returns job or subjob object
        logger.debug("TagPrepareLCGRTHandler prepare called, %s", job.id)

        # prepare inputdata
        input_files = []
        input_guids = []
       
        if job.inputdata:

            # check for subjobs
            if job._getRoot().subjobs:
                if job.inputdata._name == 'ATLASLocalDataset':
                    if not job.inputdata.names: raise ApplicationConfigurationError(None,'No inputdata has been specified.')
                    input_files = job.inputdata.names

                elif job.inputdata._name == 'DQ2Dataset':
                    if not job.inputdata.names: raise ApplicationConfigurationError(None,'No inputdata has been specified. Failure in job %s.%s. Dataset %s' %(job._getRoot().id, job.id, job.inputdata.dataset)  )
                    input_guids = job.inputdata.guids
                    input_files = job.inputdata.names
                    job.inputdata.type = 'DQ2_COPY'

            else:
                if job.inputdata._name == 'ATLASLocalDataset':
                    input_files = ATLASLocalDataset.get_filenames(app)

                elif job.inputdata._name == 'DQ2Dataset':
                    
                    job.inputdata.type = 'DQ2_COPY'
                    input_guids, input_files = _splitlist(job.inputdata.get_contents())
                    job.inputdata.names = input_files          
                    job.inputdata.guids = input_guids          

        if job.outputdata:
            raise ApplicationConfigurationError(None,'No outputdata required for TagPrepare job.')

        if job._getRoot().subjobs:
            jid = "%d.%d" % (job._getRoot().id, job.id)
        else:
            jid = "%d" % job.id

        if getConfig('LCG')['JobLogHandler'] == 'DQ2':
            raise ApplicationConfigurationError(None,'Staging of log files in DQ2 requested but not possible for TagPrepare.')
        
        # prepare inputsandbox
        inputbox = [File(os.path.join(__athdirectory__,'athena-utility.sh')),
                    File(os.path.join(__directory__,'get_tag_info.py'))]
        if input_guids:     _append_file_buffer(inputbox,'input_guids',input_guids)
        if input_files:     _append_file_buffer(inputbox,'input_files',input_files)

        exe = os.path.join(__directory__,'run-tagprepare-lcg.sh')
        outputbox = jobmasterconfig.outputbox
        requirements = jobmasterconfig.requirements.__copy__()
        environment  = jobmasterconfig.env.copy()
        
        # If ArgSplitter is used
        try:
            if job.application.args:
                environment['ATHENA_OPTIONS'] = environment['ATHENA_OPTIONS'] + ' ' + ' '.join(job.application.args)
                if job.application.options:
                    job.application.options = job.application.options + ' ' + job.application.args
                else:
                    job.application.options = job.application.args
        except AttributeError:
            pass
        
        output_location = ''
        environment['OUTPUT_LOCATION'] = output_location
        environment['ATLASOutputDatasetLFC'] = config['ATLASOutputDatasetLFC']

        # Fix DATASETNAME env variable for DQ2_COPY mode
        if job.inputdata and job.inputdata._name == 'DQ2Dataset':
            if job.inputdata.dataset:
                from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import resolve_container
                datasets = resolve_container(job.inputdata.dataset)
                environment['DATASETNAME'] = datasets[0]
                try:
                    environment['DATASETLOCATION'] = ':'.join(job.inputdata.get_locations(overlap=False)[ datasets[0] ])
                except:
                    printout = 'Job submission failed ! Dataset %s could not be found in DQ2 ! Maybe retry ?' %(datasets[0])
                    raise ApplicationConfigurationError(None,printout )

        # Work around for glite WMS spaced environement variable problem
        inputbox.append(FileBuffer('athena_options',environment['ATHENA_OPTIONS']+'\n'))

        # append a property for monitoring to the jobconfig of subjobs
        lcg_config = LCGJobConfig(File(exe), inputbox, [], outputbox, environment, [], requirements)
        return lcg_config

    def master_prepare( self, app, appconfig):
        """Prepare the master job"""

        job = app._getParent() # Returns job or subjob object
        logger.debug('TagPrepareLCGRTHandler master_prepare called: %s', job.id )

        self.username = gridProxy.identity(safe=True)

        # Check if all sites are in the same cloud
        if job.backend.requirements.sites:
            firstCloud = whichCloud(job.backend.requirements.sites[0])
            for site in job.backend.requirements.sites:
                cloud = whichCloud(site)
                if cloud != firstCloud:
                    printout = 'Job submission failed ! Site specified with j.backend.requirements.sites=%s are not in the same cloud !' %(job.backend.requirements.sites)
                    raise ApplicationConfigurationError(None,printout )


        # prepare input sandbox
        inputbox = [ ( File(os.path.join(__athdirectory__,'athena-utility.sh')) ),
                     ( File(os.path.join(__directory__,'get_tag_info.py')))]
            
        # CN: added TNTJobSplitter clause  
        if job.inputdata and job.inputdata._name == 'DQ2Dataset':
            _append_files(inputbox,os.path.join(__athdirectory__, 'ganga-stage-in-out-dq2.py'),
                          os.path.join(__athdirectory__, 'dq2_get'),
                          os.path.join(__athdirectory__, 'dq2info.tar.gz'))

        ## insert more scripts to inputsandbox for FileStager
        if job.inputdata and job.inputdata._name == 'DQ2Dataset' and job.inputdata.type in ['FILE_STAGER']:
            _append_files(inputbox,'make_filestager_joption.py','dm_util.py','fs-copy.py')
            #_append_files(inputbox,'make_filestager_joption.py','dm_util.py')

        #       add libDCache.so and libRFIO.so to fix broken access in athena 12.0.x
        if not 'ganga-stage-in-out-dq2.py' in [ os.path.basename(file.name) for file in inputbox ]:
            _append_files(inputbox, os.path.join(__athdirectory__, 'ganga-stage-in-out-dq2.py'))
        if not 'dq2tracerreport.py' in [ os.path.basename(file.name) for file in inputbox ]:
            _append_files(inputbox, os.path.join(__athdirectory__,'dq2tracerreport.py'))
        if not 'db_dq2localid.py' in [ os.path.basename(file.name) for file in inputbox ]:
            _append_files(inputbox, os.path.join(__athdirectory__, 'db_dq2localid.py'))
        if not 'getstats.py' in [ os.path.basename(file.name) for file in inputbox ]:
            _append_files(inputbox, os.path.join(__athdirectory__, 'getstats.py'))

        _append_files(inputbox,os.path.join(__athdirectory__, 'libdcap.so'))

        if job.inputsandbox: inputbox += job.inputsandbox
            
        # prepare environment
        environment={
            'MAXNUMREFS'     : str(app.max_num_refs),
            'STREAM_REF'     : app.stream_ref,
            'ATLAS_RELEASE'  : app.atlas_release,
            'ATHENA_OPTIONS' : '',
            'ATHENA_USERSETUPFILE' : '',
            'ATLAS_PROJECT' : '',
            'ATLAS_EXETYPE' : 'ATHENA',
            'GANGA_VERSION' : configSystem['GANGA_VERSION']
        }

        environment['DCACHE_RA_BUFFER'] = config['DCACHE_RA_BUFFER']
        requirements = AtlasLCGRequirements()
        
        if job.inputdata and job.inputdata._name == 'DQ2Dataset':
            if job.inputdata.dataset:
                datasetname = job.inputdata.dataset
                environment['DATASETNAME'] = ':'.join(datasetname)
                environment['DATASETLOCATION'] = ':'.join(job.inputdata.get_locations())
                environment['DQ2_URL_SERVER'] = configDQ2['DQ2_URL_SERVER']
                environment['DQ2_URL_SERVER_SSL'] = configDQ2['DQ2_URL_SERVER_SSL']
                environment['DATASETTYPE'] = job.inputdata.type
                if job.inputdata.failover:
                    environment['DATASETFAILOVER'] = 1
                environment['DATASETDATATYPE'] = job.inputdata.datatype
                if job.inputdata.accessprotocol:
                    environment['DQ2_LOCAL_PROTOCOL'] = job.inputdata.accessprotocol
                if job.inputdata.check_md5sum:
                    environment['GANGA_CHECKMD5SUM'] = 1
                    
            else:
                raise ApplicationConfigurationError(None,'j.inputdata.dataset is empty - DQ2 dataset name needs to be specified.')

            # Raise submission exception
            if (not job.backend.CE and 
                not (job.backend.requirements._name == 'AtlasLCGRequirements' and job.backend.requirements.sites) and
                not (job.splitter and job.splitter._name == 'DQ2JobSplitter') and
                not (job.splitter and job.splitter._name == 'TNTJobSplitter') and
                not (job.splitter and job.splitter._name == 'AnaTaskSplitterJob')):

                raise ApplicationConfigurationError(None,'Job submission failed ! Please use DQ2JobSplitter or specify j.backend.requirements.sites or j.backend.requirements.CE !')

            if job.inputdata.match_ce_all or job.inputdata.min_num_files>0:
                raise ApplicationConfigurationError(None,'Job submission failed ! Usage of j.inputdata.match_ce_all or min_num_files is obsolete ! Please use DQ2JobSplitter or specify j.backend.requirements.sites or j.backend.requirements.CE !')
            #if job.inputdata.number_of_files and (job.splitter and job.splitter._name == 'DQ2JobSplitter'):
            #    allLoc = job.inputdata.get_locations(complete=0)
            #    completeLoc = job.inputdata.get_locations(complete=1)
            #    incompleteLoc = []
            #    for loc in allLoc:
            #        if loc not in completeLoc:
            #            incompleteLoc.append(loc)
            #    if incompleteLoc:
            #        raise ApplicationConfigurationError(None,'Job submission failed ! Dataset is incomplete ! Usage of j.inputdata.number_of_files and DQ2JobSplitter is not allowed for incomplete datasets !')

        # prepare job requirements
        cmtconfig = app.atlas_cmtconfig
        if not cmtconfig in ['i686-slc4-gcc34-opt', 'i686-slc5-gcc43-opt']:
            cmtconfig = 'i686-slc4-gcc34-opt'

        requirements.software = ['VO-atlas-offline-%s-%s' %(app.atlas_release, cmtconfig )]

        #       add software requirement of dq2clients
        if job.inputdata and job.inputdata.type in [ 'DQ2_DOWNLOAD', 'TNT_DOWNLOAD', 'DQ2_COPY', 'FILE_STAGER']:
            try:
                # override the default one if the dq2client_version is presented 
                # in the job backend's requirements object
                dq2client_version = job.backend.requirements.dq2client_version
            except AttributeError:
                pass
            if dq2client_version:
                requirements.software += ['VO-atlas-dq2clients-%s' % dq2client_version]
                environment['DQ2_CLIENT_VERSION'] = dq2client_version

#       jobscript

        exe = os.path.join(__directory__,'run-tagprepare-lcg.sh')
        #exe = os.path.join(__directory__,'get_tag_info.py')

#       output sandbox
        outputbox = [
            'taginfo.pkl'
        ]


        if job.outputsandbox: outputbox += job.outputsandbox

        return LCGJobConfig(File(exe),inputbox,[],outputbox,environment,[],requirements) 

from Ganga.GPIDev.Credentials import GridProxy
gridProxy = GridProxy()

allHandlers.add('TagPrepare','LCG',TagPrepareLCGRTHandler)

config = getConfig('Athena')
configDQ2 = getConfig('DQ2')
configSystem = getConfig('System')
logger = getLogger()



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: GridSandboxCache.py,v 1.10 2009-07-16 10:41:17 hclee Exp $
###############################################################################
#
# LCG backend
#
# ATLAS/ARDA
#
# Date:   January 2007
import re

from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Base.Proxy import getName
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from Ganga.GPIDev.Credentials import getCredential

from Ganga.Utility.logging import getLogger
from Ganga.Lib.LCG.Utility import get_uuid

from Ganga.Utility.ColourText import ANSIMarkup, NoMarkup, Foreground, Effects


class GridFileIndex(GangaObject):

    '''
    Data object for indexing a file on the grid. 

    @author: Hurng-Chun Lee 
    @contact: hurngchunlee@gmail.com
    '''

    _schema = Schema(Version(1, 0), {
        'id': SimpleItem(defvalue='', doc='the main identity of the file'),
        'name': SimpleItem(defvalue='', doc='the name of the file'),
        'md5sum': SimpleItem(defvalue='', doc='the md5sum of the file'),
        'attributes': SimpleItem(defvalue={}, doc='a key:value pairs of file metadata')
    })

    _category = 'GridFileIndex'
    _name = 'GridFileIndex'

    logger = getLogger()

    def __init__(self):
        super(GridFileIndex, self).__init__()

    def __eq__(self, other):
        return other.id == self.id


class GridSandboxCache(GangaObject):

    '''
    Helper class for upladong/downloading/deleting sandbox files on a grid cache. 

    @author: Hurng-Chun Lee 
    @contact: hurngchunlee@gmail.com
    '''

    _schema = Schema(Version(1, 1), {
        'vo': SimpleItem(defvalue='dteam', hidden=1, copyable=0, doc='the Grid virtual organization'),
        'middleware': SimpleItem(defvalue='EDG', hidden=1, copyable=1, doc='the LCG middleware type'),
        'protocol': SimpleItem(defvalue='', copyable=1, doc='file transfer protocol'),
        'max_try': SimpleItem(defvalue=1, doc='max. number of tries in case of failures'),
        'timeout': SimpleItem(defvalue=180, copyable=0, hidden=1, doc='transfer timeout in seconds'),
        'uploaded_files': ComponentItem('GridFileIndex', defvalue=[], sequence=1, protected=1, copyable=0, hidden=1, doc='a repository record for the uploaded files')
    })

    _category = 'GridSandboxCache'
    _name = 'GridSandboxCache'
    _exportmethods = ['upload', 'download', 'delete',
                      'get_cached_files', 'list_cached_files', 'cleanup']

    logger = getLogger()

    def __init__(self):
        super(GridSandboxCache, self).__init__()

    def upload(self, files=[], opts=''):
        """
        Uploads multiple files to a remote grid storage.

        @param files is a list of local files to be uploaded to the grid.
               The elemement can be a file path or a File object.

        @return True if files are successfully uploaded; otherwise it returns False
        """
        status = False

        paths = []
        for f in files:
            if getName(f) == 'File':
                paths.append('file://%s' % f.name)
            elif getName(f) == 'str':
                paths.append('file://%s' % f)
            else:
                self.logger.warning('unknown file expression: %s' % repr(f))

        uploaded_files = self.impl_upload(files=paths, opts=opts)

        if len(uploaded_files) == len(files):
            status = self.impl_bookkeepUploadedFiles(
                uploaded_files, append=True, opts=opts)
        else:
            status = False

        if len(uploaded_files) == len(files):
            status = self.impl_bookkeepUploadedFiles(
                uploaded_files, append=True, opts=opts)
        else:
            status = False

        return status

    def download(self, files=[], dest_dir=None, opts=''):
        """
        Downloads multiple files from remote grid storages to 
        a local directory.

        If the file is successfully downloaded, the local file path would be:

            - os.path.join(dest_dir, os.path.basename(local_fname_n)

        @param files is a list of files to be downloaded from the grid.
               The data format of it should be:
               - [index_grid_file_1, index_grid_file_2, ...]

        @param dest_dir is a local destination directory to store the downloaded files.

        @return True if files are successfully downloaded; otherwise it returns False
        """
        status = False
        myFiles = self.__get_file_index_objects__(files)
        downloadedFiles = self.impl_download(
            files=myFiles, dest_dir=dest_dir, opts=opts)

        if len(downloadedFiles) == len(myFiles):
            status = True
        else:
            self.logger.warning('some files not successfully downloaded')

        return status

    def delete(self, files=[], opts=''):
        """
        Deletes multiple files from remote grid storages.

        @param files is a list of files to be deleted from the grid.
               The data format of it should be:
               - [index_grid_file_1, index_grid_file_2, ...]

        @return True if files are successfully deleted; otherwise it returns False
        """
        status = False
        myFiles = self.__get_file_index_objects__(files)
        deletedFiles = self.impl_delete(files=myFiles, opts=opts)

        if len(deletedFiles) == len(myFiles):
            status = True
        else:
            self.logger.warning('some files not successfully deleted')

        return status

    def cleanup(self, opts=''):
        """
        Cleans up the uploaded files.

        @return True if all grid files listed in the index file are successfully deleted.
        """
        status = False

        all_files = self.get_cached_files()

        f_ids = []
        for f in all_files:
            f_ids.append(f.id)

        return self.delete(files=f_ids)

    def get_cached_files(self, opts=''):
        """
        Gets the indexes of the uploaded files on the grid. 

        @return the dictionary indexing the uploaded files on the grid.
                The key of the dictionary should be the main index (e.g. GUID) of the grid files.
        """
        return self.impl_getUploadedFiles(opts=opts)

    def list_cached_files(self, loop=True, opts=''):
        """
        Lists the uploaded files.

        if loop = True, it prints also the uploaded files associated with subjobs.
        """

        fc = 0
        ds = ''

        doColoring = True

        fg = Foreground()
        fx = Effects()

        status_colors = {'inuse': fg.orange,
                         'free': fg.blue,
                         'gone': fg.red}

        status_mapping = {'new': 'inuse',
                          'submitted': 'inuse',
                          'submitting': 'inuse',
                          'running': 'inuse',
                          'completed': 'free',
                          'completing': 'free',
                          'failed': 'free',
                          'killed': 'free'}

        if doColoring:
            markup = ANSIMarkup()
        else:
            markup = NoMarkup()

        def __markup_by_status__(fileIndex, counter, status):

            fmtStr = '\n%4d\t%-30s\t%-12s\t%s' % (
                counter, fileIndex.name, status, fileIndex.id)

            try:
                return markup(fmtStr, status_colors[status])
            except KeyError:
                return markup(fmtStr, fx.normal)

        j = self.getJobObject()

        for f in self.get_cached_files(opts=opts):

            my_status = 'unknown'

            if j:
                try:
                    my_status = status_mapping[j.status]
                except KeyError:
                    pass

            ds += __markup_by_status__(f, fc, my_status)

            fc += 1

        if j and loop:
            for sj in j.subjobs:
                for f in sj.backend.sandboxcache.get_cached_files(opts=opts):

                    my_status = 'unknown'

                    try:
                        my_status = status_mapping[sj.status]
                    except KeyError:
                        pass

                    ds += __markup_by_status__(f, fc, my_status)

                    fc += 1

        return ds

    # methods to be implemented in the child classes
    def impl_upload(self, files=[], opts=''):
        """
        Uploads multiple files to a remote grid storage.

        @param files is a list of files in URL format (i.e. file://...)

        @return a list of successfully uploaded files represented by GridFileIndex objects
        """
        raise NotImplementedError

    def impl_download(self, files=[], dest_dir=None, opts=''):
        """
        Downloads multiple files from remote grid storages to 
        a local directory.

        @param files is a list of files represented by GridFileIndex objects 
        @param dest_dir is a local destination directory to store the downloaded files.

        @return a list of successfully downloaded files represented by GridFileIndex objects
        """
        raise NotImplementedError

    def impl_delete(self, files=[], opts=''):
        """
        Deletes multiple files from remote grid storages. 

        @param files is a list of files represented by GridFileIndex objects 
        @return a list of successfully deleted files represented by GridFileIndex objects
        """
        raise NotImplementedError

    def impl_bookkeepUploadedFiles(self, files=[], append=True, opts=''):
        """
        basic implementation for bookkeeping the uploaded files.
        It simply keeps the GridFileIndex objects in the job repository.

        @param files is a list of files represented by GridFileIndex objects 
        @return True if files are successfully logged in the local index file 
        """

        self.uploaded_files = files

        return True

    def impl_getUploadedFiles(self, opts=''):
        """
        basic implementation for getting the previously uploaded files from the
        job repository.

        @return a list of files represented by GridFileIndex objects
        """
        files = self.uploaded_files

        return files

    # private methods
    def __get_file_index_objects__(self, files=[]):
        '''Gets file index object according to the given file list
             - try to get the GridFileIndex object from the local index file.  

        @param files is a list of file indexes
        @return a list of files represented by GridFileIndex objects
        '''

        cachedFiles = self.get_cached_files()
        myFiles = []
        for f in cachedFiles:
            if f.id in files:
                myFiles.append(f)

        return myFiles

    def __get_unique_fname__(self):
        '''gets an unique filename'''
        cred = getCredential('GridProxy', self.middleware)
        uid = re.sub(r'[\:\-\(\)]{1,}', '', cred.identity()).lower()
        fname = 'user.%s.%s' % (uid, get_uuid())
        return fname

    def __cmd_retry_loop__(self, shell, cmd, maxRetry=3):
        '''Executing system command with retry feature'''
        i = 0
        rc = 0
        output = None
        m = None
        try_again = True
        while try_again:
            i = i + 1
            self.logger.debug('run cmd: %s' % cmd)
            rc, output, m = shell.cmd1(cmd, allowed_exit=[0, 255])
            if rc in [0, 255]:
                try_again = False
            elif i == maxRetry:
                try_again = False
            else:
                self.logger.warning("trial %d: error: %s" % (i, output))

        return (rc, output, m)



#!/usr/bin/env python
from __future__ import print_function

import os,os.path,shutil,tempfile
import sys,time
import glob
import sys

# FIXME: print as DEBUG: to __syslog__ file
#print(sys.path)
#print(os.environ['PATH'])
#print(sys.version)

# bugfix #13314 : make sure that the wrapper (spawned process) is detached from Ganga session
# the process will not receive Control-C signals
# using fork  and doing setsid() before  exec would probably  be a bit
# better (to avoid  slim chance that the signal  is propagated to this
# process before setsid is reached)
# this is only enabled if the first argument is 'subprocess' in order to enable
# running this script by hand from outside ganga (which is sometimes useful)
if len(sys.argv)>1 and sys.argv[1] == 'subprocess':
    os.setsid()

###########################################################################################

###INLINEMODULES###

############################################################################################

input_sandbox = ###INPUT_SANDBOX###
sharedoutputpath= ###SHAREDOUTPUTPATH###
outputpatterns = ###OUTPUTPATTERNS###
appscriptpath = ###APPSCRIPTPATH###
environment = ###ENVIRONMENT###
workdir = ###WORKDIR###

statusfilename = os.path.join(sharedoutputpath,'__jobstatus__')

try:
    statusfile=open(statusfilename,'w')
except IOError as x:
    print('ERROR: not able to write a status file: ', statusfilename)
    print('ERROR: ',x)
    raise

line='START: '+ time.strftime('%a %b %d %H:%M:%S %Y',time.gmtime(time.time())) + os.linesep
statusfile.writelines(line)
statusfile.flush()

if not os.path.exists(workdir):
    os.makedirs(workdir)
os.chdir(workdir)

##-- WARNING: get the input files including the python modules BEFORE sys.path.insert()
# -- SINCE PYTHON 2.6 THERE WAS A SUBTLE CHANGE OF SEMANTICS IN THIS AREA

for f in input_sandbox:
    getPackedInputSandbox(f)

# -- END OF MOVED CODE BLOCK

#get input files
###DOWNLOADINPUTFILES###

# create inputdata list
###CREATEINPUTDATALIST###

import sys
sys.path.insert(0, ###GANGADIR###)
sys.path.insert(0,os.path.join(os.getcwd(),PYTHON_DIR))

import subprocess
import tarfile

fullenvironment = os.environ.copy()
for key,value in environment.iteritems():
    fullenvironment[key] = value

outfile=open('stdout','w')
errorfile=open('stderr','w')

sys.stdout=open('./__syslog__','w')
sys.stderr=sys.stdout

import subprocess

import time #datetime #disabled for python2.2 compatiblity

try:
    child = subprocess.Popen(appscriptpath, shell=False, stdout=outfile, stderr=errorfile, env=fullenvironment)
except OSError as x:
    errfile = open( 'tt', 'w' )
    errfile.close()
    print('EXITCODE: %d'%-9999, file=statusfile)
    print('FAILED: %s'%time.strftime('%a %b %d %H:%M:%S %Y'), file=statusfile) #datetime.datetime.utcnow().strftime('%a %b %d %H:%M:%S %Y')
    print('PROBLEM STARTING THE APPLICATION SCRIPT: %s %s'%(appscriptpath,str(x)), file=statusfile)
    statusfile.close()
    sys.exit()

print('PID: %d'%child.pid, file=statusfile)
statusfile.flush()

result = -1

try:
    while 1:
        result = child.poll()
        if result is not None:
            break
        outfile.flush()
        errorfile.flush()
        time.sleep(0.3)
finally:
    pass

    sys.stdout=sys.__stdout__
    sys.stderr=sys.__stderr__


outfile.flush()
errorfile.flush()

createOutputSandbox(outputpatterns,None,sharedoutputpath)

def printError(message):
    errorfile.write(message + os.linesep)
    errorfile.flush()

def printInfo(message):
    outfile.write(message + os.linesep)
    outfile.flush()


###OUTPUTUPLOADSPOSTPROCESSING###

outfile.close()
errorfile.close()

###OUTPUTSANDBOXPOSTPROCESSING###

line="EXITCODE: " + repr(result) + os.linesep
line+='STOP: '+time.strftime('%a %b %d %H:%M:%S %Y',time.gmtime(time.time())) + os.linesep
statusfile.writelines(line)
statusfile.close()
sys.exit()




from __future__ import absolute_import
from .LCG import LCG, LCGJobConfig
from .CREAM import CREAM
from .ARC import ARC
from .LCGRequirements import LCGRequirements
from .GridSandboxCache import GridFileIndex, GridSandboxCache
from .LCGSandboxCache import LCGFileIndex, LCGSandboxCache
from .GridftpSandboxCache import GridftpFileIndex, GridftpSandboxCache
from .ElapsedTimeProfiler import ElapsedTimeProfiler
from .LCGOutputDownloader import LCGOutputDownloadTask, LCGOutputDownloadAlgorithm, LCGOutputDownloader



import os
import sys
import time
import datetime
import glob
import pickle
## NB parseCommandLine first then import Dirac!!
from DIRAC.Core.Base.Script import parseCommandLine
parseCommandLine()
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
dirac = Dirac()

# Write to output pipe
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
#reserved_stdout = sys.stdout
#sys.stdout = sys.stderr
# def output(object):
#    reserved_stdout.write("CHECK ME " + str(object))
#    reserved_stdout.write(pickle.dumps(object))

# Dirac commands
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

def getJobGroupJobs(jg): output(dirac.selectJobs(jobGroup=jg))


def kill(id): output(dirac.delete(id))


def peek(id): output(dirac.peek(id))


def getJobCPUTime(id): output(dirac.getJobCPUTime(id))


def reschedule(id): output(dirac.reschedule(id))


def submit(djob, mode='wms'): output(dirac.submit(djob, mode=mode))


def ping(system, service): output(dirac.ping(system, service))


def removeFile(lfn):
    ret = {}
    if type(lfn) is list:
        for l in lfn:
            ret.update(dirac.removeFile(l))
    else:
        ret.update(dirac.removeFile(lfn))
    output(ret)


def getMetadata(lfn): output(dirac.getMetadata(lfn))


def getReplicas(lfns): output(dirac.getReplicas(lfns))


def getFile(lfns, destDir=''): output(dirac.getFile(lfns, destDir=destDir))


def replicateFile(lfn, destSE, srcSE, locCache=''):
    res = dirac.replicateFile(lfn, destSE, srcSE, locCache)
    output(res)
    sys.stdout.write(res)


def removeReplica(lfn, sE):
    output(dirac.removeReplica(lfn, sE))


def getOutputData(id, outputFiles='', destinationDir=''):
    output(dirac.getJobOutputData(id, outputFiles, destinationDir))


def splitInputData(files, files_per_job):
    output(dirac.splitInputData(files, files_per_job))


def getInputDataCatalog(lfns, site, xml_file):
    output(dirac.getInputDataCatalog(lfns, site, xml_file))


def uploadFile(lfn, file, diracSEs, guid=None):
    outerr = {}
    for se in diracSEs:
        result = dirac.addFile(lfn, file, se, guid)
        if result.get('OK', False) and lfn in result.get('Value', {'Successful': {}})['Successful']:
            result['Value']['Successful'][lfn].update({'DiracSE': se})
            md = dirac.getMetadata(lfn)
            if md.get('OK', False) and lfn in md.get('Value', {'Successful': {}})['Successful']:
                guid = md['Value']['Successful'][lfn]['GUID']
                result['Value']['Successful'][lfn].update({'GUID': guid})
            output(result)
            return
        outerr.update({se: result})
    else:
        output(outerr)


def addFile(lfn, file, diracSE, guid):
    output(dirac.addFile(lfn, file, diracSE, guid))


def getOutputSandbox(id, outputDir=os.getcwd(), oversized=True):
    result = dirac.getOutputSandbox(id, outputDir, oversized)
    if result is not None and result.get('OK', False):
        tmpdir = os.path.join(outputDir, str(id))
        os.system('mv -f %s/* %s/.' % (tmpdir, outputDir))
        os.system('rm -rf %s' % tmpdir)
        ganga_logs = glob.glob('%s/*_Ganga_*.log' % outputDir)

        if ganga_logs:
            os.system('ln -s %s %s/stdout' % (ganga_logs[0], outputDir))
    output(result)


def getOutputDataInfo(id):
    ret = {}
    result = getOutputDataLFNs(id, pipe_out=False)
    if result.get('OK', False) and 'Value' in result:
        for lfn in result.get('Value', []):
            file_name = os.path.basename(lfn)
            ret.update({file_name: {'LFN': lfn}})
            md = dirac.getMetadata(lfn)
            if md.get('OK', False) and lfn in md.get('Value', {'Successful': {}})['Successful']:
                ret[file_name].update(
                    {'GUID': md['Value']['Successful'][lfn]['GUID']})
            # this catches if fail upload, note lfn still exists in list as
            # dirac tried it
            elif md.get('OK', False) and lfn in md.get('Value', {'Failed': {}})['Failed']:
                ret[file_name].update({'LFN': '###FAILED###'})
                ret[file_name].update(
                    {'LOCATIONS': md['Value']['Failed'][lfn]})
                ret[file_name].update({'GUID': 'NotAvailable'})
                continue
            rp = dirac.getReplicas(lfn)
            if rp.get('OK', False) and lfn in rp.get('Value', {'Successful': {}})['Successful']:
                ret[file_name].update(
                    {'LOCATIONS': rp['Value']['Successful'][lfn].keys()})
    output(ret)


# could shrink this with dirac.getJobOutputLFNs from ##dirac
def getOutputDataLFNs(id, pipe_out=True):
    parameters = dirac.parameters(id)
    lfns = []
    ok = False
    message = 'The outputdata LFNs could not be found.'

    if parameters is not None and parameters.get('OK', False):
        parameters = parameters['Value']
        # remove the sandbox if it has been uploaded
        sandbox = None
        if 'OutputSandboxLFN' in parameters:
            sandbox = parameters['OutputSandboxLFN']

        # now find out about the outputdata
        if 'UploadedOutputData' in parameters:
            lfn_list = parameters['UploadedOutputData']
            import re
            lfns = re.split(',\s*', lfn_list)
            if sandbox is not None and sandbox in lfns:
                lfns.remove(sandbox)
            ok = True
        elif parameters is not None and 'Message' in parameters:
            message = parameters['Message']

    result = {'OK': ok}
    if ok:
        result['Value'] = lfns
    else:
        result['Message'] = message

    if pipe_out:
        output(result)
    return result


def normCPUTime(id):
    parameters = dirac.parameters(id)
    ncput = None
    if parameters is not None and parameters.get('OK', False):
        parameters = parameters['Value']
        if 'NormCPUTime(s)' in parameters:
            ncput = parameters['NormCPUTime(s)']
    output(ncput)


def status(job_ids):
    # Translate between the many statuses in DIRAC and the few in Ganga
    statusmapping = {'Checking': 'submitted',
                     'Completed': 'completed',
                     'Deleted': 'failed',
                     'Done': 'completed',
                     'Failed': 'failed',
                     'Killed': 'killed',
                     'Matched': 'submitted',
                     'Received': 'submitted',
                     'Running': 'running',
                     'Staging': 'submitted',
                     'Stalled': 'running',
                     'Waiting': 'submitted'}

    result = dirac.status(job_ids)
    if not result['OK']:
        output(result)
        return
    status_list = []
    bulk_status = result['Value']
    for id in job_ids:
        job_status = bulk_status.get(id, {})
        minor_status = job_status.get('MinorStatus', None)
        dirac_status = job_status.get('Status', None)
        dirac_site = job_status.get('Site', None)
        ganga_status = statusmapping.get(dirac_status, None)
        if ganga_status is None:
            ganga_status = 'failed'
            dirac_status = 'Unknown: No status for Job'
        if dirac_status == 'Completed' and (minor_status not in ['Pending Requests']):
            ganga_status = 'running'
        if minor_status in ['Uploading Output Data']:
            ganga_status = 'running'

        try:
            from DIRAC.Core.DISET.RPCClient import RPCClient
            monitoring = RPCClient('WorkloadManagement/JobMonitoring')
            app_status = monitoring.getJobAttributes(
                id)['Value']['ApplicationStatus']
        except:
            app_status = "unknown ApplicationStatus"

        status_list.append([minor_status, dirac_status, dirac_site,
                            ganga_status, str(app_status)])

    output(status_list)


def getStateTime(id, status):
    log = dirac.loggingInfo(id)
    if 'Value' not in log:
        output(None)
        return
    L = log['Value']
    checkstr = ''

    if status == 'running':
        checkstr = 'Running'
    elif status == 'completed':
        checkstr = 'Done'
    elif status == 'completing':
        checkstr = 'Completed'
    elif status == 'failed':
        checkstr = 'Failed'
    else:
        checkstr = ''

    if checkstr == '':
        sys.stdout.write(None)
        return

    for l in L:
        if checkstr in l[0]:
            T = datetime.datetime(
                *(time.strptime(l[3], "%Y-%m-%d %H:%M:%S")[0:6]))
            output(T)
            return
    output(None)


def timedetails(id):
    log = dirac.loggingInfo(id)
    d = {}
    for i in range(0, len(log['Value'])):
        d[i] = log['Value'][i]
    output(d)

# DiracAdmin commands
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/


def getJobPilotOutput(id, dir):
    pwd = os.getcwd()
    try:
        os.chdir(dir)
        os.system('rm -f pilot_%d/std.out' % id)
        os.system('rmdir pilot_%d' % id)
        result = DiracAdmin().getJobPilotOutput(id)
    finally:
        os.chdir(pwd)
    output(result)


def getServicePorts(): output(DiracAdmin().getServicePorts())


def getSitesForSE(se):
    from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE
    result = getSitesForSE(se)
    output(result)


def getSEsForSite(site):
    from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
    result = getSEsForSite(site)
    output(result)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
from GangaGaudi.Lib.RTHandlers.RunTimeHandlerUtils import get_share_path, master_sandbox_prepare, sandbox_prepare, script_generator
from GangaDirac.Lib.RTHandlers.DiracRTHUtils import dirac_inputdata, dirac_ouputdata, mangle_job_name, diracAPI_script_template, diracAPI_script_settings, API_nullifier, dirac_outputfile_jdl
from GangaDirac.Lib.Files.DiracFile import DiracFile
from Ganga.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns, getWNCodeForOutputPostprocessing
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Lib.File import File, FileBuffer
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from Ganga.Utility.util import unique
from Ganga.GPIDev.Base.Proxy import isType, stripProxy
logger = getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class ExeDiracRTHandler(IRuntimeHandler):

    """The runtime handler to run plain executables on the Dirac backend"""

    def master_prepare(self, app, appmasterconfig):
        inputsandbox, outputsandbox = master_sandbox_prepare(app, appmasterconfig)
        if type(app.exe) == File:
            input_dir = app.getJobObject().getInputWorkspace().getPath()
            exefile = os.path.join(input_dir, os.path.basename(app.exe.name))
            if not os.path.exists(exefile):
                msg = 'Executable: "%s" must exist!' % str(exefile)
                raise ApplicationConfigurationError(None, msg)

            os.system('chmod +x %s' % exefile)
        return StandardJobConfig(inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):
        inputsandbox, outputsandbox = sandbox_prepare(app, appsubconfig, appmasterconfig, jobmasterconfig)
        input_data,   parametricinput_data = dirac_inputdata(app)
#        outputdata,   outputdata_path      = dirac_ouputdata(app)

        job = stripProxy(app).getJobObject()
        outputfiles = [this_file for this_file in job.outputfiles if isType(this_file, DiracFile)]

        commandline = app.exe
        if isType(app.exe, File):
            #logger.info("app: %s" % str(app.exe.name))
            #fileName = os.path.join(get_share_path(app), os.path.basename(app.exe.name))
            #logger.info("EXE: %s" % str(fileName))
            #inputsandbox.append(File(name=fileName))
            inputsandbox.append(app.exe)
            commandline = os.path.basename(app.exe.name)
        commandline += ' '
        commandline += ' '.join([str(arg) for arg in app.args])
        logger.debug('Command line: %s: ', commandline)

        #exe_script_path = os.path.join(job.getInputWorkspace().getPath(), "exe-script.py")
        exe_script_name = 'exe-script.py'

        inputsandbox.append(FileBuffer(name=exe_script_name,
                            contents=script_generator(exe_script_template(),
                                                    #remove_unreplaced = False,
                                                    # ,
                                                    COMMAND=commandline,
                                                    OUTPUTFILESINJECTEDCODE = getWNCodeForOutputPostprocessing(job, '    ')
                                                    ),
                                       executable=True))

        dirac_outputfiles = dirac_outputfile_jdl(outputfiles)

        # NOTE special case for replicas: replicate string must be empty for no
        # replication
        dirac_script = script_generator(diracAPI_script_template(),
                                        DIRAC_IMPORT='from DIRAC.Interfaces.API.Dirac import Dirac',
                                        DIRAC_JOB_IMPORT='from DIRAC.Interfaces.API.Job import Job',
                                        DIRAC_OBJECT='Dirac()',
                                        JOB_OBJECT='Job()',
                                        NAME=mangle_job_name(app),
                                        # os.path.basename(exe_script_path),
                                        EXE=exe_script_name,
                                        # ' '.join([str(arg) for arg in app.args]),
                                        EXE_ARG_STR='',
                                        EXE_LOG_FILE='Ganga_Executable.log',
                                        ENVIRONMENT=None,  # app.env,
                                        INPUTDATA=input_data,
                                        PARAMETRIC_INPUTDATA=parametricinput_data,
                                        OUTPUT_SANDBOX=API_nullifier(outputsandbox),
                                        OUTPUTFILESSCRIPT=dirac_outputfiles,
                                        OUTPUT_PATH="",  # job.fqid,
                                        OUTPUT_SE=getConfig('DIRAC')['DiracOutputDataSE'],
                                        SETTINGS=diracAPI_script_settings(app),
                                        DIRAC_OPTS=job.backend.diracOpts,
                                        REPLICATE='True' if getConfig('DIRAC')['ReplicateOutputData'] else '',
                                        # leave the sandbox for altering later as needs
                                        # to be done in backend.submit to combine master.
                                        # Note only using 2 #s as auto-remove 3
                                        INPUT_SANDBOX='##INPUT_SANDBOX##'
                                        )


        #logger.info("inbox: %s" % str(unique(inputsandbox)))
        #logger.info("outbox: %s" % str(unique(outputsandbox)))

        return StandardJobConfig(dirac_script,
                                 inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def exe_script_template():
    script_template = """#!/usr/bin/env python
'''Script to run Executable application'''
from os import system, environ, pathsep, getcwd
import sys

# Main
if __name__ == '__main__':

    environ['PATH'] = getcwd() + (pathsep + environ['PATH'])        
    rc = (system('###COMMAND###')/256)

    ###OUTPUTFILESINJECTEDCODE###
    sys.exit(rc)
"""
    return script_template

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
allHandlers.add('Executable', 'Dirac', ExeDiracRTHandler)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
"""Merges DST files."""
import commands
import inspect
import os
import string
import subprocess
import tempfile
from Ganga.GPIDev.Adapters.IPostProcessor import PostProcessException
from Ganga.GPIDev.Adapters.IMerger import IMerger
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Base.Proxy import GPIProxyObject
from Ganga.GPIDev.Schema import ComponentItem, FileItem, Schema
from Ganga.GPIDev.Schema import SimpleItem, Version
from Ganga.GPIDev.Lib.File import File
from Ganga.Utility.Config import makeConfig, ConfigError, getConfig
from Ganga.Utility.Plugin import allPlugins
from Ganga.Utility.logging import getLogger, log_user_exception

logger = getLogger()


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

class LHCbFileMerger(IMerger):

    """A merger object for LHCb files

    The merger uses DaVinci to combine DST files that have
    been returned *locally* in a job's outputsandbox. As such
    it is mainly useful for microDST files.

    The usage is as with other merger objects. See the help for 
    TextMerger or RootMerger for more details.

    Example:

    lhcbm = LHCbFileMerger()
    lhcbm.files = ['dv.dst']

    This object can be attached to a job object or 
    used to merge a list of jobs with its merge 
    method.

    It is possible to overide the default opts file
    for performing the merge. A new opts file can 
    be provided via the 'merge_opts' field. This should
    be done with care, as some opts are assumed when
    writing the files for output.

    Comments and help requests should be sent to:

    lhcb-distributed-analysis@cern.ch
    """

    _category = 'postprocessor'
    _exportmethods = ['merge']
    _name = 'LHCbFileMerger'
    _schema = IMerger._schema.inherit_copy()
    docstr = 'The version of DaVinci to use when merging. (e.g. v33r0p1)'
    _schema.datadict['version'] = SimpleItem(defvalue='', doc=docstr)

    def mergefiles(self, file_list, output_file):

        # if no opts file is specified, then use version from installation

        # this is the bit specifing the files
        output_opts = """
outputfile = '%s' """ % output_file
        output_opts += "\ninput = ["
        file_sep = ','
        for f in file_list:
            if f is file_list[-1]:
                file_sep = ''
            output_opts += "'%s' %s " % (f, file_sep)
        output_opts += "]"

        output_opts += """
from GaudiConf import IOHelper

IOHelper().inputFiles(input)
IOHelper().outStream(outputfile,"InputCopyStream")

from Configurables import LHCbApp

LHCbApp().EvtMax = -1
        """

        # write this out to a file
        opts_file_name = tempfile.mktemp('.py')
        opts_file = file(opts_file_name, 'w')
        try:
            opts_file = file(opts_file_name, 'w')
            opts_file.write(output_opts)
        finally:
            opts_file.close()

        if not os.path.exists(opts_file_name):
            msg = "Failed to write temporary options file '%s' during merge"
            raise PostProcessException(msg % opts_file_name)

        import EnvironFunctions
        script_file_name = EnvironFunctions.construct_merge_script(self.version,
                                                                   opts_file_name)

        return_code = subprocess.call(['/bin/sh', script_file_name])
        if return_code != 0:
            msg = 'The LHCbFileMerger returned %i when calling gaudirun'
            logger.warning(msg % return_code)

        # finally clean up
        os.unlink(script_file_name)
        os.unlink(opts_file_name)

        if not os.path.exists(output_file):
            msg = "The output file '%s' was not created"
            raise PostProcessException(msg % output_file)
        # needed as exportmethods doesn't seem to cope with inheritance

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

# Add it to the list of plug-ins
allPlugins.add(LHCbFileMerger, 'postprocessor', 'LHCbFileMerger')

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



# Import classes that should be in the Tasks namespace
## Import the list of tasks and the task and abstract job definition

## Tasks
from MCTask import MCTask
from AnaTask import AnaTask
from MultiTask import MultiTask

## Transforms
from MCTransforms import EvgenTransform, SimulTransform, ReconTransform
from AnaTransform import AnaTransform
from MultiTransform import MultiTransform
from AtlasTask import AtlasTask
from AtlasTransform import AtlasTransform
from AtlasUnit import AtlasUnit

# Applications
from TaskApplication import AthenaTask, AthenaMCTask, AthenaMCTaskSplitterJob, AnaTaskSplitterJob

# Start Logger
#import Ganga.Utility.logging
#logger = Ganga.Utility.logging.getLogger()
from Ganga.GPIDev.Lib.Tasks.common import logger



#! /usr/bin/env python

import os, sys

from getopt import getopt,GetoptError
from threading import Thread
from commands import getstatusoutput

from lfc import *

def usage():

    print 'Name:'
    print '    ganga-stagein.py'
    print
    print 'Arguments:'
    print '    logical names'
    print 
    print 'Options:'
    print '    -h, --help            this prinout'
    print '    -i, --input file      list of logical names'
    print '    -d, --directory path  to stage the input files (default $PWD)'
    print '    -t, --timeout seconds for the staging in (default 900)'
    print '    -r, --retry number    for the staging command (default 3)'
    print '    -v, --verbose         verbosity'

def get_guid(lfn):
    '''Get guid for a lfn
    '''

    statg = lfc_filestatg()
    rc = lfc_statg(lfn,'',statg)
    if not rc: return statg.guid

def get_replicas(lfn):
    '''List replicas and sort the one on close SE first
    '''

    replicas = []

    listp = lfc_list()
    res = lfc_listreplica(lfn,'',CNS_LIST_BEGIN,listp)
    while res:
        if res.host in closeSEs:
            replicas.insert(0,res.sfn)
        else:
            replicas.append(res.sfn)
        res = lfc_listreplica(lfn,'',CNS_LIST_CONTINUE,listp)

    lfc_listreplica(lfn,'',CNS_LIST_END,listp)

    return replicas
    
class PoolFileCatalog:
    '''Helper class to create PoolFileCatalog.xml
    '''
    def __init__(self,name='PoolFileCatalog.xml'):

        self.pfc = open(name,'w')
        print >>self.pfc,'<?xml version="1.0" ?>'
        print >>self.pfc,'<POOLFILECATALOG>'

    def addFile(self,guid,lfn,pfn):

        print >>self.pfc,'    <File ID="%s">' % guid
        print >>self.pfc,'        <logical>'
        print >>self.pfc,'            <lfn name="%s"/>' % lfn
        print >>self.pfc,'        </logical>'
        print >>self.pfc,'        <physical>'
        print >>self.pfc,'            <pfn filetype="ROOT_All" name="%s"/>' % pfn
        print >>self.pfc,'        </physical>'
        print >>self.pfc,'    </File>'

    def close(self):

        print >>self.pfc,'</POOLFILECATALOG>'

class StageIn(Thread):

    def __init__(self,lfn,replicas,file):

        Thread.__init__(self)
        self.lfn = lfn
        self.replicas = replicas
        self.file = file

    def run(self):

        for rep in self.replicas:
            for r in xrange(0,retry):
                if  verbose: print 'INFO LFN: %s Replica: %s Retry: %d' % (lfn,rep,r)
                cmd = 'lcg-cp --vo atlas -t %d %s file:%s' % (timeout,rep,self.file)
                rc, out = getstatusoutput(cmd)
                if not rc: return

                print 'Return code %d from %s' % (rc,cmd)
                print out



if __name__ == '__main__':

    directory = os.getcwd()
    retry = 2
    timeout = 900
    input = None
    verbose = False

    try:
        opts, args = getopt(sys.argv[1:],'ht:d:r:i:v',['help','directory=','input=','timeout=','retry=','verbose'])
    except GetoptError:
        usage()
        sys.exit(1)

    for opt, val in opts:

        if opt in ['-h','--help']:
            usage()
            sys.exit()

        if opt in ['-d','--directory']:
            direcory = val 

        if opt in ['-i','--input']:
            input = val

        if opt in ['-t','--timeout']:
            timeout = int(val)

        if opt in ['-r','--retry']:
            retry = int(val)

        if opt in ['-v','--verbose']:
            verbose = True

    if input:
        lfns = [ line.strip() for line in file(input) ]
    else:
        lfns = args

    if not len(lfns):
        print 'No files requested.'
        sys.exit()


#   determine the closeSEs

    rc, output = getstatusoutput('edg-brokerinfo getCloseSEs')
    if rc:
        print 'ERROR: Could not determine close SEs'
        closeSEs = []
    else:
        closeSEs = output.split()
        print 'INFO: Close SEs are ' + ', '.join(closeSEs)

    pfc = PoolFileCatalog()
    workers=[]
    try: lfc_startsess('','')
    except NameError: pass

    for lfn in lfns:

        if verbose: print 'LFN: %s' % lfn
        guid = get_guid(lfn)
        if not guid:
            print 'ERROR: LFN %s not found.' % lfn 
            continue
        if verbose: print 'GUID: %s' % guid

        name = os.path.basename(lfn)
        pfn = os.path.join(directory,name)

        pfc.addFile(guid,name,pfn)

        replicas = get_replicas(lfn)
        if not replicas:
            print 'ERROR: No replica found for LFN %s' % lfn
            continue

        if verbose:
            print 'Replicas :\n   %s' % '\n   '.join(replicas)   

        s = StageIn(lfn,replicas,pfn)
        s.start()
        workers.append(s)

    pfc.close()

    try: lfc_stopsess()
    except NameError: pass

    for s in workers:
        s.join()



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: StagerJobSplitter.py,v 1.2 2008-12-11 20:37:10 hclee Exp $
###############################################################################
# Athena StagerJobSplitter

import math
import os.path

from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Adapters.ISplitter import ISplitter
from Ganga.GPIDev.Lib.Job import Job
from Ganga.GPIDev.Schema import *
from Ganga.Lib.Mergers import RootMerger 

from Ganga.Utility.logging import getLogger

logger = getLogger()

class StagerJobSplitter(ISplitter):
    '''Job splitting for StagerDataset'''

    _name     = 'StagerJobSplitter'
    _category = 'splitters'
    _schema   = Schema(Version(1,0), {
        'numfiles' : SimpleItem(defvalue=0,doc='Number of files per subjob'),
        'scheme'   : SimpleItem(defvalue='local', doc='The job splitting scheme: \'local\', \'lcg\'', hidden=1),
    })

    _GUIPrefs = [ { 'attribute' : 'numfiles', 'widget' : 'Int' } ]
    
    def __make_subjob__(self, mj, guids, names, sjob_evnts=-1, sites=None):
        
        """
        private method to create subjob object
        """
        
        logger.debug('generating subjob to run %d events in-total on files: %s' % (sjob_evnts, repr(guids)))
        j = Job()

        j.name            = mj.name
        j.inputdata       = mj.inputdata

        if j.inputdata.type in ['','DQ2']:
            j.inputdata.guids = guids

        j.inputdata.names = names

        j.outputdata    = mj.outputdata
        j.application   = mj.application
        if sjob_evnts != -1:
            j.application.max_events = sjob_evnts
        j.backend       = mj.backend
        
        if j.backend._name in ['LCG'] and j.backend.requirements._name == 'AtlasLCGRequirements':
            if sites:
                j.backend.requirements.sites = sites
        
        j.inputsandbox  = mj.inputsandbox
        j.outputsandbox = mj.outputsandbox

        return j

    def split(self,job):

        logger.debug('StagerJobSplitter called')

        if job.inputdata._name != 'StagerDataset':
            raise ApplicationConfigurationError(None,'StagerJobSplitter requires StagerDataset as input')

        if job.application._name not in  ['Athena']:
            raise ApplicationConfigurationError(None,'StagerJobSplitter requires Athena as application')

        if self.numfiles <= 0: 
            self.numfiles = 1

        total_evnts = -1
        try:
            total_evnts = int(job.application.max_events)
        except ValueError:
            pass

        sjob_evnts  = -1

        subjobs = []

        ## resolve pfns and guids
        pfns = job.inputdata.get_surls()
        myguids = pfns.keys()

        myguids.sort()

        mynames = []
        for guid in myguids:
            mynames.append( pfns[guid] )

        if job.inputdata.type in ['', 'DQ2']:
            job.inputdata.guids = myguids

        job.inputdata.names = mynames

        nrjob = int(math.ceil(len(mynames)/float(self.numfiles)))

        if total_evnts != -1:
            sjob_evnts = int(math.ceil( total_evnts/float(nrjob) ))

        ## split scheme for local jobs: simple splitting
        if self.scheme.lower() == 'local': 
            for i in xrange(0,nrjob):
                subjobs.append( self.__make_subjob__(job, myguids[i*self.numfiles:(i+1)*self.numfiles], mynames[i*self.numfiles:(i+1)*self.numfiles], sjob_evnts) )

        return subjobs



"""Demonstration IAction implementations.

The DemoExtractor and DemoReporter classes provide simple examples of how to
reuse Core actions adding to the extracted data and summary reports.

This example is referenced in the WIKI page:
http://twiki.cern.ch/twiki/bin/view/ArdaGrid/HowToGangaRobot

See DEMO.INI for suitable configuration options.

"""

from GangaRobot.Lib.Base.BaseExtractor import BaseExtractor
from GangaRobot.Lib.Base.BaseReporter import BaseReporter
from GangaRobot.Lib.Core.CoreExtractor import CoreExtractor
from GangaRobot.Lib.Core.CoreReporter import CoreReporter
import random

class DemoExtractor(BaseExtractor):
    """Simple example of how to reuse CoreExtractor, adding to the extracted data."""
    
    def __init__(self):
        """Add CoreExtractor and self to chain of extractors."""
        self.chain = [CoreExtractor(), self]

    def handlejobnode(self, jobnode, job):
        """Add a 'demo' subnode to each 'job' node with a random integer value in range [0,100]."""
        dn = jobnode.addnode('demo')
        dn.addnode('value', random.randint(0,100))

class DemoReporter(BaseReporter):
    """Simple example of how to reuse CoreReporter, adding to the summary report."""
    
    def __init__(self):
        """Add CoreReporter and self to chain of reporters."""
        self.chain = [CoreReporter(), self]
    
    def handlereport(self, report, runnode):
        """Set the report title and add a line reporting the average of the demo values."""
        report.title = 'Demo Report'
        values = runnode.getvalues('job.demo.value')
        total = 0
        for value in values:
            total += int(value)
        avg = total / max(len(values), 1)
        report.addline('Average value: %d' % avg)



"""Action interface.

The IAction interface should be implemented by actions to be executed by Driver.

"""

from GangaRobot.Framework import Utility
from Ganga.Utility.Config import ConfigError

class IAction(object):
    
    """Action interface used by Driver to execute run actions.
    
    Implementations should override the execute() method, and must provide a
    zero-argument constructor if overriding __init__(). 
    
    Typical actions might be Submitter, Extractor, Reporter, etc.

    See GangaRobot.Lib.Base for abstract implementations, and
    GangaRobot.Lib.Core for concrete implementations.
    
    N.B. Driver creates a new instance of the IAction class for each execution,
    so modifying instance attributes in the execute method will not affect other
    executions of the same action.
    
    N.B. Implementations can use the getoption() method if they want to allow
    users the possibility to override configuration options programmatically for
    a given instance.
     
    """
    
    def execute(self, runid):
        """Action specific execute method.
        
        Keyword arguments:
        runid -- A UTC ID string which identifies the run.
         
        This method is called by Driver in the context of a Ganga session.
        
        """
        raise NotImplementedError
        
    def getoption(self, key):
        """Return the configuration option value.
        
        If the instance has an attribute 'options' containing the given key,
        then the corresponding value is returned, otherwise the value is
        retrieved from the [Robot] section of the global configuration.
        
        This provides a single point of access to configuration options and
        allows users the possibility to override options programmatically for a
        given instance.
        
        Example:
        c = CoreFinisher()
        c.options = {'BaseFinisher_Timeout':3600}
        
        """
        if hasattr(self, 'options') and key in self.options:
            return self.options[key]
        else:
            try:        
                return Utility.getconfig()[key]
            except ConfigError:         
                return ''



##########################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: GridShell.py,v 1.1 2008-07-17 16:41:00 moscicki Exp $
#
# Copyright (C) 2003 The Ganga Project
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# LCG backend
#
# ATLAS/ARDA
#
# Date:   August 2006
##########################################################################

import os

from Ganga.Utility.Shell import Shell
from Ganga.Utility.Config import getConfig, ConfigError
from Ganga.Utility.logging import getLogger

_allShells = {}


def getShell(middleware='EDG', force=False):
    """
    Utility function for getting Grid Shell.
    Caller should take responsiblity of credential checking if proxy is needed.

    Argumennts:

     middleware - grid m/w used 
     force      - False : if the shell already exists in local cache return the previous created instance
                  True  : recreate the shell and if not None update the cache
    """

    logger = getLogger()

    if not middleware:
        logger.debug('No middleware specified, assuming default EDG')
        middleware = 'EDG'

    if middleware in _allShells.keys() and not force:
        return _allShells[middleware]

    values = {}
    for key in ['X509_USER_PROXY', 'X509_CERT_DIR', 'X509_VOMS_DIR']:
        try:
            values[key] = os.environ[key]
        except KeyError:
            pass

    configname = ""
    if middleware == 'EDG' or middleware == 'GLITE':
        configname = 'LCG'
    else:
        configname = middleware

    config = None
    try:
        config = getConfig(configname)
    except:
        logger.warning(
            '[%s] configuration section not found. Cannot set up a proper grid shell.' % configname)
        return None

    s = None

    key = '%s_SETUP' % middleware

    # 1. check if the *_SETUP is changed by user -> take the user's value as session value
    # 2. else check if *_LOCATION is defined as env. variable -> do nothing (ie. create shell without any lcg setup)
    # 3. else take the default *_SETUP as session value

    MIDDLEWARE_LOCATION = '%s_LOCATION' % middleware

    if config.getEffectiveLevel(key) == 2 and MIDDLEWARE_LOCATION in os.environ:
        s = Shell()
    else:
        if os.path.exists(config[key]):
            # FIXME: Hardcoded rule for ARC middleware setup (pass explicitly
            # the $ARC_LOCATION as argument), this is hardcoded to maintain
            # backwards compatibility (and avoid any side effects) for EDG and
            # GLITE setup scripts which did not take any arguments
            if key.startswith('ARC') and MIDDLEWARE_LOCATION in os.environ:
                s = Shell(
                    config[key], setup_args=[os.environ[MIDDLEWARE_LOCATION]])
            else:
                s = Shell(config[key])
        else:
            logger.warning("Configuration of %s for %s: " %
                           (middleware, configname))
            logger.warning("File not found: %s" % config[key])

    if s:
        for key, val in values.items():
            s.env[key] = val

        # check and set env. variables for default LFC setup
        if 'LFC_HOST' not in s.env:
            try:
                s.env['LFC_HOST'] = config['DefaultLFC']
            except ConfigError:
                pass

        if 'LFC_CONNTIMEOUT' not in s.env:
            s.env['LFC_CONNTIMEOUT'] = '20'

        if 'LFC_CONRETRY' not in s.env:
            s.env['LFC_CONRETRY'] = '0'

        if 'LFC_CONRETRYINT' not in s.env:
            s.env['LFC_CONRETRYINT'] = '1'

        _allShells[middleware] = s

    return s



"""
Provides Persistency and the base class for all Registries
Also, a list of all Registries is kept here
"""

from Ganga.Core.GangaRepository.GangaRepository import GangaRepository, RepositoryError, InaccessibleObjectError, SchemaVersionError
from Ganga.Core.GangaRepository.Registry import RegistryError, RegistryAccessError, RegistryKeyError, RegistryLockError, ObjectNotInRegistryError
from Ganga.Core.GangaRepository import GangaRepositoryXML

allRegistries = {}


def addRegistry(registry):
    allRegistries[registry.name] = registry


def getRegistries():
    return allRegistries.values()


def getRegistry(name):
    return allRegistries[name]



#! /usr/bin/env python
#
# determine SE type
#

import os, urllib, sys, imp, commands, re, socket

# error codes
# WRAPLCG_UNSPEC
EC_UNSPEC        = 410000
# WRAPLCG_WNCHECK_UNSPEC
EC_Configuration = 410100
# WRAPLCG_STAGEIN_UNSPEC
EC_QueryFiles    = 410300
# WRAPLCG_STAGEIN_LCGCP
EC_DQ2GET        = 410302
# WRAPLCG_WNCHECK_PROXY
EC_PROXY         = 410101
# WRAPLCG_STAGEOUT_UNSPEC
EC_STAGEOUT      = 410400


########################################################################
def getTiersOfATLASCache():
    """Download TiersOfATLASCache.py"""
    
    url = 'http://atlas.web.cern.ch/Atlas/GROUPS/DATABASE/project/ddm/releases/TiersOfATLASCache.py'
    local = os.path.join(os.environ['PWD'],'TiersOfATLASCache.py')
    try:
        urllib.urlretrieve(url, local)
    except IOError:
        print 'Failed to download TiersOfATLASCache.py'
        
    try:
        tiersofatlas = imp.load_source('',local)
    except SyntaxError:
        print 'Error loading TiersOfATLASCache.py'
        sys.exit(EC_UNSPEC)

    return tiersofatlas

##########################
def findsetype(sitesrm):
    setype= 'NULL'
    
    if sitesrm.find('castor')>=0:
        setype = 'CASTOR'
    elif sitesrm.find('dpm')>=0:
        setype = 'DPM'
    elif sitesrm.find('pnfs')>=0:
        setype = 'DCACHE'
    elif sitesrm.find('/nfs/')>=0:
        setype = 'NFS'

    return setype

##########################
def localsitename():
    # Determine local domainname
    hostname = None
    domainname = None
    # First choice: EDG_WL_RB_BROKERINFO or GLITE_WMS_RB_BROKERINFO
    if 'EDG_WL_RB_BROKERINFO' in os.environ:
        try:
            f = open(os.environ['EDG_WL_RB_BROKERINFO'], "r")
            lines = f.readlines()
            for line in lines:
                match = re.search('name = "(\S*):2119', line)
                if match:
                    hostname =  [ match.group(1) ]
        except:
            pass

    if 'GLITE_WMS_RB_BROKERINFO' in os.environ:
        try:
            f = open(os.environ['GLITE_WMS_RB_BROKERINFO'], "r")
            lines = f.readlines()
            for line in lines:
                match = re.search('name = "(\S*):2119', line)
                if match:
                    hostname =  [ match.group(1) ]
        except:
            pass

    # Second choice: GANGA_LCG_CE
    if not hostname and 'GANGA_LCG_CE' in os.environ:
        try:
            hostname = re.findall('(\S*):2119',os.environ['GANGA_LCG_CE'])
            lcgcename = re.findall('(\S*):2119',os.environ['GANGA_LCG_CE'])[0]
            #print hostname, lcgcename
        except:
            pass

    # Third choice: VO_ATLAS_DEFAULT_SE
    if not hostname and 'VO_ATLAS_DEFAULT_SE' in os.environ:
        hostname = os.environ['VO_ATLAS_DEFAULT_SE']
        if hostname.find('grid.sara.nl')>=0: hostname = ''
            
    # Fourth choice: local hostname
    if not hostname:
        hostname = socket.gethostbyaddr(socket.gethostname())

    if hostname.__class__.__name__=='list' or hostname.__class__.__name__=='tuple':
        hostname = hostname[0]

    domainname = re.sub('^[\w\-]+\.','',hostname)
    if domainname=='gridka.de':
        domainname = 'fzk.de'

    tiersofatlas = getTiersOfATLASCache()
    localsites = []
    localsiteseq = []

    # See if local domainname is in TiersOfAtlasCache.py
    for site, desc in tiersofatlas.sites.iteritems():
        sitename = desc['domain'].strip('.*')
        if sitename=='':
            continue
        try:
            srm = desc['srm'].strip()
        except KeyError:
            continue
        pat1 = re.compile(sitename)
        found1 = re.findall(pat1,domainname)
        pat2 = re.compile(domainname)
        found2 = re.findall(pat2,sitename)
        if sitename == domainname: 
            localsites.append(site)
            localsiteseq.append(site)
        elif found1 or found2:
            localsites.append(site)

    # Get location list of dataset
    try:
        datasetlocation = os.environ['DATASETLOCATION'].split(":")
    except:
        #print "ERROR : DATASETLOCATION not defined"
        datasetlocation = []
        pass

    # Choose local site configuration (if possible eq datasetlocation)
    localsite = ''
    # 1st pick: if domainname eq sitename(ToA)  
    for site in localsiteseq:
        localsite = site
        if site in datasetlocation:
            break
    # 2nd pick: if domainname is in sitename(ToA) or the other way around  
    if not localsite:
        for site in localsites:
            localsite = site
            if site in datasetlocation:
                break

    localsiteid = localsite 
    # change CERNCAF to CERN
    localsiteid = re.sub('^CERNCAF$','CERN',localsiteid)
    # change TIER0 to CERN
    localsiteid = re.sub('^TIER0','CERN',localsiteid)
    # remove TAPE
    localsiteid = re.sub('TAPE$','',localsiteid)
    # remove DISK
    localsiteid = re.sub('DISK$','',localsiteid)
    # remove PANDA
    localsiteid = re.sub('PANDA$','',localsiteid)

    return localsiteid

##########################
if __name__ == '__main__':

    setype = 'NULL'
    
    tiersofatlas = getTiersOfATLASCache()

    if 'VO_ATLAS_DEFAULT_SE' in os.environ:
        sitese=os.environ['VO_ATLAS_DEFAULT_SE']
        #print 'ganga_setype.py: VO_ATLAS_DEFAULT_SE=%s'%sitese
    elif sys.argv[1:]:
        sitese=sys.argv[1]
    else:
        print 'ERROR no SE specified !'
        sys.exit(4)

    # Find TiersOfAtlasCache entry
    if sitese:
        for site, desc in tiersofatlas.sites.iteritems():
            try:
                sitesrm = desc['srm'].strip()
            except KeyError:
                continue
            if sitesrm.find(sitese)>=0:
                setype = findsetype(sitesrm)

            if setype == 'NULL':
                sitesrm_domain = re.sub('^[\w\-]+\.','',sitesrm)
                sitese_domain = re.sub('^[\w\-]+\.','',sitese)
                if sitesrm_domain.find(sitese_domain)>=0:
                    setype = findsetype(sitesrm_domain)


    # If not in TiersOfAtlasCache
    if setype == 'NULL':
        cmd = 'lcg-info --list-se --query SE=%s --attr Accesspoint --sed 2>/dev/null' %sitese
        rc, out = commands.getstatusoutput(cmd)
        out2 = out.split('%')
        if len(out2)>1:
            setype = findsetype(out2[1])

    localsiteid = localsitename()
    if localsiteid == 'RAL':
        setype = 'DCACHE'
    if localsiteid == 'NIKHEF':
        setype = 'DPM'
    
        
    print setype
    



#!/usr/bin/env python

# GuidExtractor.py - pulls out GUIDs from an event collection 
# and finds corresponding logical filenames.
# Copyright (c) 2006 C.Nicholson, M. Kenyon.  All rights reserved.
# For licence conditions please read LICENCE file.

import xml.dom
import xml.dom.minidom
import getopt
import os
import re
import sys
import commands
import random
import popen2
from Ganga.Utility.logging import getLogger
logger = getLogger()

try:
   import lfc
except:
   logger.debug('Error importing lfc module')
   pass

try:
   from dq2.clientapi.DQ2 import DQ2
   from dq2.common.Config import Config
   from dq2.common.DQException import *
   from dq2.location.client.LocationClient import LocationClient
except ImportError:
   logger.warning("Environment not set [error importing DQ2 dependencies]!")
   logger.warning("Try setting PYTHONPATH to the dq2-client directory.")
   pass

import GenerateCatalogs

actions = {
   'test'   : [0, False],
   'query'   : [0, False]
}

# global maps to be stored
_guidLfnMap = {}
_guidDatasetMap = {}


def chkargs(command, args):
   """Check action and number of arguments
   """
   try:
      # arguments are args minus action
      nargs = len(args)
      # get number of mandatory args and whether there is a limit
      nm = actions[command][0]
      nolimit = actions[command][1]
      # check if number of args is correct
      if nargs < nm or (nargs > nm and not nolimit):
         return False
      if nolimit and actions[command][2] and (len(args)-nm)%2 == 1:
         logger.warning('Error: you must give matching pairs of arguments')
         return False
   except:
      return False

   return True



def usage():
   """
   Usage: will go here   """
   logger.info(usage.__doc__)



def pythonCalled(opts = []):
   
   generateMiniCats = False
   generatePOOLcat = False
   for o in opts:
     if o == 'm':
       generateMiniCats = True
     if o == 'p':
       generatePOOLcat = True
     if o != 'm' and o != 'p':
       #srcRootFile = o
       src_collection_name = o

   #return main(generateMiniCats, generatePOOLcat, srcRootFile)
   return main(generateMiniCats, generatePOOLcat, src_collection_name)



def standAlone(argv):

   generateMiniCats = False
   generatePOOLcat = False
   if len(argv) < 1:
      usage()
      return
   logger.info("length of args = " + str(len(argv)))
   logger.info("ARGS =" + str(argv))

   try:
      opts, args = getopt.getopt(argv[0:], "mp")
   except getopt.GetoptError:
      logger.info("Invalid arguments!")
      usage()
      return

   for o, a in opts:
      if o == '-m':
         generateMiniCats = True
      if o == '-p':
         generatePOOLcat = True

   return main(generateMiniCats, generatePOOLcat)

def getGuidDatasetMap():
   #print _guidDatasetMap
   return _guidDatasetMap


def getGuidLfnMap():
   #print _guidLfnMap
   return _guidLfnMap

def writeCollListFileGUIDXML(src_collection_name):
      #print "########################"
      #print "IN writeCollListFileGUIDXML - again using tag file as source and not myEvent created by CollAppend"
      #print "########################"
      collListfile = open("CollListFileGUID.exe.xml","w")
      collListfile.write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>\n")
      collListfile.write("<!DOCTYPE ArgList>\n")
      collListfile.write("<ArgList>\n")
      collListfile.write("<ToolInfo date=\"Thu Nov 20 18:29:03 2008&#xA;\" toolID=\"CollListFileGUID.exe\">\n")
      #collListfile.write("<CliArg name=\"local\" option=\"-xmlInput\">ThisCondition &gt; ThisNumber</CliArg>\n")
      collListfile.write("<CliArg name=\"SrcInfo\" option=\"-src\">"+src_collection_name+" RootCollection"+"</CliArg>\n")
      #collListfile.write("<CliArg name=\"SrcInfo\" option=\"-src\">"+ srcRootFile +" RootCollection"+"</CliArg>\n")
      collListfile.write("</ToolInfo>\n")
      collListfile.write("</ArgList>\n")


#def main(generateMiniCats, generatePOOLcat, srcRootFile):
def main(generateMiniCats, generatePOOLcat, src_collection_name):

   guids = []
   outputLFNs = False
   outputGUIDs = False
   #print "########################"
   #print "IN GUIDEXTRACTOR MAIN"
   #print "########################"
   
   #does CollListFileGUID  exist in the current directory?
   cwd = os.getcwd()
   if not os.path.exists("CollListFileGUID.exe..xml"):
      writeCollListFileGUIDXML(src_collection_name)
   else:
      writeCollListFileGUIDXML(src_collection_name) 

   try:
      dq = DQ2 (
            con_url = Config().getConfig('dq2-content-client').get('dq2-content-client', 'insecure'),
            con_urlsec = Config().getConfig('dq2-content-client').get('dq2-content-client', 'secure'),
                                                                                                                                           
            loc_url = Config().getConfig('dq2-location-client').get('dq2-location-client', 'insecure'),
            loc_urlsec = Config().getConfig('dq2-location-client').get('dq2-location-client', 'secure'),
                                                                                                                                           
            rep_url = Config().getConfig('dq2-repository-client').get('dq2-repository-client', 'insecure'),
            rep_urlsec = Config().getConfig('dq2-repository-client').get('dq2-repository-client', 'secure'),
                                                                                                                                           
            sub_url = Config().getConfig('dq2-subscription-client').get('dq2-subscription-client', 'insecure'),
            sub_urlsec = Config().getConfig('dq2-subscription-client').get('dq2-subscription-client', 'secure'),
      )
         
     
      guid_string = commands.getoutput("CollListFileGUID.exe -xmlInput CollListFileGUID.exe.xml |\
                                   grep -E [[:alnum:]]{8}'-'[[:alnum:]]{4}'-'[[:alnum:]]{4}'-'[[:alnum:]]{4}'-'[[:alnum:]]{12} " \
                                  )
      #print guid_string
      ##########################################################
      if guid_string.find('Exception') > -1:
         logger.warning("Error: incorrect query, or problem with collection utilities")
         return
 
      guids = guid_string.split()
      if len(guids) == 0:
         logger.warning("Error: query returned no files")
         return

      guids = guid_string.split()

      # get file and dataset name by doing GUID lookup in DQ2
      dataset_names = []
      vuids = []
      guidDatasetMap = {}
      guidLfnMap = {}
      files = []
      firstLoop = True

      for guid in guids:
         if guid != "StreamRAW_ref" and guid != "StreamESD_ref" and guid != "Token":
            vuid = dq.contentClient.queryDatasetsWithFileByGUID(guid)
            if len(vuid) == 0:
               logger.warning("Error: guid "+ guid + " returned by query is not registered in any DQ2 dataset!")
               logger.warning("Skipping to next file...")
               continue
            else: 
               #TODO: choose 'best' dataset, not just the first one 
               n=0
               noDataset = 'False'
               try:
                  dataset = dq.repositoryClient.resolveVUID(vuid[n])
                  name = dataset.get('dsn')
               except:
                  if (len(vuid)==1):
                     noDataset = 'True'
                     continue
            # hack to avoid non-official datasets -- FIXME!!!
               match = re.compile('user')
               while (match.search(name) and n < len(vuid)-1):
                  try:
                     dataset = dq.repositoryClient.resolveVUID(vuid[n+1])
                     name = dataset.get('dsn')
                     noDataset = 'False'
                  except:
                     noDataset = 'True'
                     n=n+1
                     continue
                  n=n+1
               if match.search(name):
                  logger.warning("No official dataset found for guid "+guid)
                  noDataset = 'True'
                  continue

               if noDataset == 'True':
                  name = 'NO.DATASET'

               if vuids.count(vuid[n]) == 0:
                  vuids.append(vuid[n])
                  dataset_names.append(name)
               guidDatasetMap[guid] = name
               files = dq.listFilesInDataset(name)
               #print "###########FILES###########"
               #print files
               #print "######################"
               try:
                  lfn = (files[0][guid])['lfn']
                  #print lfn
                  guidLfnMap[guid] = str(lfn)
               except:
                  logger.warning("Could not resolve GUID "+guid+" to any file in dataset "+str(name))
                  guidDatasetMap[guid] = 'NO.DATASET'

         global _guidDatasetMap
         #print "guidDatasetMap:"
         #print _guidDatasetMap
         global _guidLfnMap
         #print "guidLfnMap:"
         #print _guidLfnMap
         _guidDatasetMap = guidDatasetMap
         _guidLfnMap = guidLfnMap
      

      # sort the list of guids by their datasets
      sorted_guids = []
      sorted_lfns = []
      i = 0
      for dataset in dataset_names:
         for item in guidDatasetMap.iteritems():
            if (item[1] == dataset):
               sorted_guids.append(item[0])                  
               sorted_lfns.append(guidLfnMap[sorted_guids[i]])
               i = i + 1

      # find location of existing files
      # information not currently used so comment out.
      """
      sites = []
      site_map = dq.locationClient.queryDatasetLocations(vuids, None)
      for ds in site_map:
         complete = site_map.get(ds).get(1)
         i = 0
         while i < int(len(complete)):
            sites.append(complete.pop())
            i = i + 1
      """

      # generate POOL XML catalogue with required files
      # if specified, also generates a 'mini-catalogue' per file
      # Should be done on-the-fly by Ganga at worker nodes, so comment out
      """ 
      if generatePOOLcat == True:
         pfns = GenerateCatalogs.getPFNsFromLFC(guidLfnMap)
         GenerateCatalogs.generateCatalog(guidLfnMap, pfns, "PoolFileCatalog.xml")
         if generateMiniCats == True:
            # for each guid, generate a POOL file catalogue just with that guid
            for guid in sorted_guids:
               miniGuidLfnMap = {}
               lfn = guidLfnMap[guid]
               miniGuidLfnMap[guid] = lfn
               name = "PoolFileCatalog_"+guid+".xml"
               GenerateCatalogs.generateCatalog(miniGuidLfnMap, pfns, name)
      """
      return guids

   except DQException as e:
      logger.warning( "Error", e)

if __name__ == '__main__':
   standAlone(sys.argv[1:])




import os
import Ganga.Utility.Config
import platform

from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

## CMSSW parameters
configCMSSW=Ganga.Utility.Config.makeConfig('CMSSW','Parameters for CMSSW')

dscrpt = 'The version CMSSW used for job submission.'
configCMSSW.addOption('CMSSW_VERSION','CMSSW_3_7_0',dscrpt)
dscrpt = 'The CMSSW setup script used for env configuration.'

config = Ganga.Utility.Config.getConfig('System')
ganga_pythonpath = config['GANGA_PYTHONPATH']

configCMSSW.addOption('CMSSW_SETUP','%s/GangaCMS/scripts/'%(ganga_pythonpath),dscrpt)
dscrpt = 'The location of the CMSSW Framework.'
configCMSSW.addOption('location','~/',dscrpt)

configMetrics = Ganga.Utility.Config.makeConfig('Metrics','List of desired metrics.')
dscrpt = 'The location of the metrics.cms list.'
configMetrics.addOption('location','%s/GangaCMS/metrics.ini'%(ganga_pythonpath),dscrpt)

dscrpt = 'The version CRAB used for job submission.'
configCMSSW.addOption('CRAB_VERSION','CRAB_2_7_5',dscrpt)

def getEnvironment( config = {} ):
    import sys
    import os.path
    import PACKAGE

    PACKAGE.standardSetup()

    arch = platform.machine()
    if not arch == 'x86_64':
        logger.error('GangaCMS> %s not supported. Different than 64 bits.'%(arch))
        return
    logger.info('GangaCMS> not supported different OS than SLC5')

    config = Ganga.Utility.Config.getConfig('CMSSW')
    cmssw_version = config['CMSSW_VERSION']
    cmssw_setup = config['CMSSW_SETUP']
    crab_version = config['CRAB_VERSION']
    cmssw_setup_script = os.path.join(cmssw_setup,'CMSSW_generic.sh')
    if not os.path.exists(cmssw_setup_script):
        logger.error('GangaCMS> CMSSW setup script not found: "%s"'%(cmssw_setup_script))
        return

    location = config['location']

    cmsswhome = os.path.join(location,cmssw_version)
    if not os.path.exists(cmsswhome):
        logger.error('GangaCMS> CMSSW location not found: "%s"'%(cmsswhome))
        return

#    from Ganga.Utility.Shell import Shell
#    shell = Shell(cmssw_setup_script)   

    logger.info('GangaCMS> [INFO] getEnvironment : done')
#    return shell.env
    return {}

def loadPlugins( config = {} ):
    import Lib.CRABTools
    import Lib.Utils
    import Lib.ConfParams

    crab_cfg_configs = {}

    for params in [Lib.ConfParams.CMSSW(),Lib.ConfParams.CRAB(),Lib.ConfParams.GRID(),Lib.ConfParams.USER()]:

      section = params.__class__.__name__
      crab_cfg_configs[section] = Ganga.Utility.Config.makeConfig('%s_CFG'%(section),'Parameters for %s at crab.cfg.'%(section))

      for k in params.schemadic.keys():
       crab_cfg_configs[section].addOption(k,None,'%s at crab.cfg'%(k))


    logger.info('GangaCMS> [INFO] loadPlugins : done')




# Example startup script for loading modules allowing Ganga submission
# of ADA jobs to DIAL backend

import os, sys
import Ganga.Utility.logging

logger = Ganga.Utility.logging.getLogger()

if logger.getEffectiveLevel() <= 20:
    print >>sys.stderr
    print >>sys.stderr, 'For help visit the ATLAS Distributed Analysis Help eGroup:'
    print >>sys.stderr, '  https://groups.cern.ch/group/hn-atlas-dist-analysis-help/'
    print >>sys.stderr

try:
    from IPython.iplib import InteractiveShell

    def magic_cmtsetup(self,args=''):
        '''Setup CMT environment.'''

        # determine the CMTHOME path
        from Ganga.Utility.Config import getConfig, ConfigError
        config = getConfig('Athena')
        cmthome = os.path.join(os.environ['HOME'], 'cmthome')
        try:
            cmthome = config['CMTHOME']
        except ConfigError:
            pass

        _do_setup(os.path.join(cmthome, 'setup.sh'), args.split())

        logger.info('CMTCONFIG = %s',os.environ['CMTCONFIG'])
        for path in os.environ['CMTPATH'].split(':'):
            logger.info('CMTPATH = %s',path)

    def magic_setup(self,args=''):
        '''Setup a CMT package'''

        import re

        _do_setup('setup.sh',args.split())

        output = os.popen('cmt -quiet show macro package').read()
        match = re.search('package=\'(.+)\'',output)
        if match:
           package = match.group(1)
           if not package == 'cmt_standalone':
               logger.info('Package %s has been configured.',package)
           else:
               logger.warning('No package directory')
        else:
           logger.error('Problem parsing cmt output')


    def magic_subjobs(self,args=''):
        '''Print status of subjobs'''

        for arg in args.split():
            try:
               jobnr = int(arg)
            except ValueError:
               logger.warning('Invalid job id %s',arg)
               continue

            logger.warning('The subjobs magic command is not supported anymore in Ganga 5')
            logger.warning('Please use instead: jobs(%s).subjobs', jobnr) 

    def magic_fixpython(self,args=''):
        '''Fix python conflict'''

        # detect the python base dir.
        _pybins = os.popen('which python').readlines()
        _pybase = sys.prefix

        if _pybins:
           _pybase = _pybins[0].split('/bin/python')[0]

        print _pybase

        _pypaths = os.environ['PYTHONPATH'].split(':')

        _new_pypaths = [] 

        # detect and remove the default python library path from PYTHONPATH
        for p in _pypaths:
            if p.find('%s/lib/python' % _pybase) >= 0:
                logger.warning('removing %s from PYTHONPATH' % p)
            else:
                _new_pypaths.append(p)

        ## reset the new python path
        os.environ['PYTHONPATH'] = ':'.join(_new_pypaths)

    InteractiveShell.magic_fixpython = magic_fixpython
    InteractiveShell.magic_cmtsetup = magic_cmtsetup
    InteractiveShell.magic_setup = magic_setup
    InteractiveShell.magic_subjobs = magic_subjobs

    del magic_cmtsetup
    del magic_setup 
    del magic_fixpython
    del InteractiveShell

except ImportError:
    pass

def _do_setup(setup,tags):

    import re

    setup=os.path.expanduser(setup)

    if not os.path.exists(setup):
       logger.warning('%s does not exist.' % setup)
       return

    if tags:
        cmd = '%s -tag=%s' % (setup,','.join(tags))
    else:
        cmd = setup

    pipe = os.popen('source %s; printenv' % cmd)
    output = pipe.read()
    rc = pipe.close()
    if rc: logger.error('non-zero return code %d from executing setup command',rc)    

    for key, value in re.findall('(\S+)=(\S+)\n',output):
        if key not in ['_','PWD','SHLVL']:
            os.environ[key] = value    



from __future__ import absolute_import
from .FileChecker import FileChecker
from .CustomChecker import CustomChecker
from .RootFileChecker import RootFileChecker
from .MetaDataChecker import MetaDataChecker



from Ganga.Core.exceptions import BackendError, ApplicationConfigurationError
from Ganga.Core.exceptions import SplitterError
from Ganga.Utility.logging import getLogger
from Ganga.Utility.Config import getConfig
from Ganga.Utility.util import unique
from GangaDirac.Lib.Splitters.SplitterUtils import DiracSplitter
from GangaDirac.Lib.Files.DiracFile import DiracFile
from Ganga.GPIDev.Base.Proxy import isType, getName, stripProxy
logger = getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def mangle_job_name(app):
    job = stripProxy(app).getJobObject()

    jobName = job.name
    jobIndex = job.getStringFQID()
    appName = str(getName(app))
    appVersion = None
    if hasattr(app, 'version'):
        appVersion = str(app.version)

    result = ''
    addBracket = False
    if jobName:
        result += '%s__' % jobName
    if appName:
        # and not j.subjobs: #not necessary?
        if not job.master and job.splitter:
            result += '{Ganga_%s_(%s.%s)' % (appName, jobIndex, '%n')
        else:
            result += '{Ganga_%s_(%s)' % (appName, jobIndex)
        addBracket = True
    if appVersion:
        result += '_%s' % appVersion
    if not result:
        result = '{Ganga_Job}'
    elif addBracket:
        result += '}'
    return result
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def API_nullifier(item):
    if item is None or len(item) == 0:
        return None
    return item

def dirac_outputfile_jdl(output_files):

    _output_files = [this_file for this_file in output_files if isType(this_file, DiracFile)]

    file_SE_dict = {}

    for this_file in _output_files:
        if not this_file.defaultSE in file_SE_dict:
            file_SE_dict[this_file.defaultSE] = []
        file_SE_dict[this_file.defaultSE].append( this_file.namePattern )

    per_SE_JDL = '''
j.setOutputData(###OUTPUTDATA###, outputPath='###OUTPUT_PATH###', outputSE=###OUTPUT_SE###)
'''
    total_JDL = ''

    for outputSE, namePatterns in file_SE_dict.iteritems():

        myLine = str(per_SE_JDL)
        myLine = myLine.replace('###OUTPUTDATA###', str(namePatterns))
        if outputSE != '':
            myLine = myLine.replace('###OUTPUT_SE###', str([outputSE]))
        else:
            myLine = myLine.replace('###OUTPUT_SE###', str([]))

        total_JDL += myLine + "\n"

    return total_JDL


def dirac_inputdata(app):
    job = stripProxy(app).getJobObject()
    input_data = None
    parametricinput_data = None

    inputLFNs = []

    if hasattr(job.inputdata, 'getLFNs'):
        inputLFNs = job.inputdata.getLFNs()

    if job.master:
        logger.debug("job.master.inputdata: %s " % str(job.master.inputdata))
    logger.debug("job.inputdata: %s" % str(job.inputdata))
    if hasattr(job.inputdata, 'getLFNs'):
        logger.debug("getLFNs(): %s" % job.inputdata.getLFNs())

    has_input_DiracFile = False
    for this_file in job.inputfiles:
        if isType(this_file, DiracFile):
            has_input_DiracFile = True
            break
    if job.master and not has_input_DiracFile:
        for this_file in job.master.inputfiles:
            if isType(this_file, DiracFile):
                has_input_DiracFile = True
                break

    if len(inputLFNs) > 0:
        # master job with a splitter reaching prepare, hence bulk submit
        if not job.master and job.splitter:
            parametricinput_data = dirac_parametric_split(app)
            if parametricinput_data is not None and len(parametricinput_data) > getConfig('DIRAC')['MaxDiracBulkJobs']:
                raise BackendError('Dirac', 'Number of bulk submission jobs \'%s\' exceeds the maximum allowed \'%s\' if more are needed please modify your config. Note there is a hard limit in Dirac of currently 1000.' % (
                    len(parametricinput_data), getConfig('DIRAC')['MaxDiracBulkJobs']))
        # master job with no splitter or subjob already split proceed as normal
        else:
            input_data = job.inputdata.getLFNs()

    elif 'Destination' not in job.backend.settings and not has_input_DiracFile:
        ##THIS IS NOT VERY DIRAC CENTRIC
        ##PLEASE WHEN TIME MOVE TO LHCBDIRAC where T1 is more applicable rcurrie
        ##Also editing the settings on the fly is asking for potential problems, should avoid
        t1_sites = getConfig('DIRAC')['noInputDataBannedSites']
        logger.info('Job has no inputdata (T1 sites will be banned to help avoid overloading them).')
        if 'BannedSites' in job.backend.settings:
            job.backend.settings['BannedSites'].extend(t1_sites)
            job.backend.settings['BannedSites'] = unique(job.backend.settings['BannedSites'])
        else:
            job.backend.settings['BannedSites'] = t1_sites[:]

    #import traceback
    # traceback.print_stack()

    return input_data, parametricinput_data

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def dirac_parametric_split(app):
    data = app.getJobObject().inputdata
    splitter = app.getJobObject().splitter

    split_data = [dataset for dataset in DiracSplitter(data, splitter.filesPerJob, splitter.maxFiles, splitter.ignoremissing)]

    split_files = []

    for dataset in split_data:
        this_dataset = []
        for this_file in dataset:
            if isType(this_file, DiracFile):
                this_dataset.append(this_file.lfn)
            else:
                raise SplitterError("ERROR: file: %s NOT of type DiracFile" % str(this_file) )
        split_files.append(this_dataset)

    if len(split_files) > 0:
        return split_files

    return None


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
def dirac_ouputdata(app):
    job = app.getJobObject()
    if job.outputdata:
        return job.outputdata.files[:], job.outputdata.location
    return None, None


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
def diracAPI_script_template():
    # NOTE setOutputData(replicate) replicate keyword only for LHCbDirac. must
    # move there when get a chance.

    import inspect
    import os.path
    script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                   'DiracRTHScript.py')

    from Ganga.GPIDev.Lib.File import FileUtils
    script_template = FileUtils.loadScript(script_location, '')

    return script_template

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def diracAPI_script_settings(app):
    job = stripProxy(app).getJobObject()
    diracAPI_line = ''
    if type(job.backend.settings) is not dict:
        raise ApplicationConfigurationError(
            None, 'backend.settings should be a dict')
    for setting, setting_val in job.backend.settings.iteritems():
        if str(setting).startswith('set'):
            _setting = str(setting)[3:]
        else:
            _setting = str(setting)
        if type(setting_val) is str:
            setting_line = 'j.set###SETTING###("###VALUE###")\n'
        else:
            setting_line = 'j.set###SETTING###(###VALUE###)\n'
        diracAPI_line += setting_line.replace('###SETTING###', _setting).replace('###VALUE###', str(setting_val))
    if diracAPI_line == '':
        diracAPI_line = None

    return diracAPI_line
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#




# XMLSummary parsing
import os, sys

if 'XMLSUMMARYBASEROOT' not in os.environ:
    sys.stderr.write("'XMLSUMMARYBASEROOT' env var not defined so summary.xml not parsed")
else:
    schemapath  = os.path.join(os.environ['XMLSUMMARYBASEROOT'],'xml/XMLSummary.xsd')
    summarypath = os.path.join(os.environ['XMLSUMMARYBASEROOT'],'python/XMLSummaryBase')
    sys.path.append(summarypath)
    import summary

    outputxml = os.path.join(os.getcwd(), 'summary.xml')
    if not os.path.exists(outputxml):
        sys.stderr.write("XMLSummary not passed as 'summary.xml' not present in working dir")
    else:
        try:
            XMLSummarydata = summary.Summary(schemapath,construct_default=False)
            XMLSummarydata.parse(outputxml)
        except:
            sys.stderr.write("Failure when parsing XMLSummary file 'summary.xml'")

        try:
            fn = open('__parsedxmlsummary__','w')
            for name, method in activeSummaryItems().iteritems():
                try:
                    fn.write( '%s = %s\n' % ( name, str(method(XMLSummarydata)) ) )
                except Exception, e:
                    fn.write( '%s = None\n' % name )
                    sys.stderr.write('XMLSummary warning: Method "%s" not available for this job\n' % name)
        except:
            sys.stderr.write('XMLSummary error: Failed to create __parsedxmlsummary__ file')
        finally:
            fn.close()

